from typing import Optional, List, Dict

from thestage.config.business.config_provider import ConfigProvider
from thestage.docker_container.dto.docker_container_create_request import DockerContainerCreateRequest
from thestage.docker_container.dto.docker_container_create_response import DockerContainerCreateResult
from thestage.docker_container.dto.docker_image_list_request import DockerImageListRequest
from thestage.docker_container.dto.docker_image_list_response import DockerImageListResponse
from thestage.global_dto.enums.order_direction_type import OrderDirectionType
from thestage.services.clients.thestage_api.core.api_client_core import TheStageApiClientCore
from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.services.clients.thestage_api.dtos.entity_filter_request import EntityFilterRequest
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList

from thestage.docker_container.dto.container_action_request import DockerContainerActionRequest
from thestage.docker_container.dto.container_response import (
    DockerContainerDto,
    DockerContainerViewResponse,
    ContainerBusinessStatusMapperResponse, DockerImageDto
)
from thestage.docker_container.dto.docker_container_list_request import DockerContainerListRequest
from thestage.docker_container.dto.docker_container_list_response import DockerContainerListResponse


class DockerContainerApiClient(TheStageApiClientCore):
    def __init__(self, config_provider: ConfigProvider):
        super().__init__(url=config_provider.get_config().main.thestage_api_url)
        self.__config_provider = config_provider

    def get_container_list(
            self,
            project_public_id: Optional[str] = None,
            project_slug: Optional[str] = None,
            statuses: List[str] = [],
            page: int = 1,
            limit: int = 10,
    ) -> PaginatedEntityList[DockerContainerDto]:
        request = DockerContainerListRequest(
            statuses=statuses,
            projectPublicId=project_public_id,
            projectSlug=project_slug,
            entityFilterRequest=EntityFilterRequest(
                orderByField="createdAt",
                orderByDirection=OrderDirectionType.DESC,
                page=page,
                limit=limit,
            ),
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/docker-container/list',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token
        )

        result = DockerContainerListResponse.model_validate(response) if response else None
        return result.paginatedList if result and result.isSuccess else None


    def get_docker_image_list(
            self,
            page: int = 1,
            limit: int = 10,
    ) -> PaginatedEntityList[DockerImageDto]:
        request = DockerImageListRequest(
            entityFilterRequest=EntityFilterRequest(
                orderByField="createdAt",
                orderByDirection=OrderDirectionType.DESC,
                page=page,
                limit=limit,
            ),
        )
        response = self._request(
            method='POST',
            url='/user-api/v1/docker-image/list',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token
        )
        result = DockerImageListResponse.model_validate(response) if response else None
        return result.paginatedList if result and result.isSuccess else None


    def get_container(
            self,
            container_slug: Optional[str] = None,
            container_public_id: Optional[str] = None,
    ) -> Optional[DockerContainerDto]:
        data = {
            "dockerContainerPublicId": container_public_id,
            "dockerContainerSlug": container_slug,
        }

        response = self._request(
            method='POST',
            url='/user-api/v2/docker-container/view',
            data=data,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        return DockerContainerViewResponse.model_validate(response).dockerContainer if response else None


    def create_container(self, request_param: DockerContainerCreateRequest) -> Optional[DockerContainerCreateResult]:
        response = self._request(
            method='POST',
            url='/user-api/v1/docker-container/create',
            data=request_param.model_dump(by_alias=True),
            token=self.__config_provider.get_config().main.thestage_auth_token
        )

        return DockerContainerCreateResult.model_validate(response) if response else None


    def container_action(
            self,
            request_param: DockerContainerActionRequest,
    ) -> TheStageBaseResponse:

        response = self._request(
            method='POST',
            url='/user-api/v2/docker-container/action',
            data=request_param.model_dump(by_alias=True),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = TheStageBaseResponse.model_validate(response) if response else None
        return result


    def get_container_business_status_map(self) -> Optional[Dict[str, str]]:
        response = self._request(
            method='POST',
            url='/user-api/v1/docker-container/status-localized-mapping',
            data=None,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        data = ContainerBusinessStatusMapperResponse.model_validate(response) if response else None
        return data.docker_container_status_map if data else None


import time
import re
from typing import Optional, List

import typer

from thestage.cli_command import CliCommand
from thestage.cli_command_helper import get_command_metadata, check_command_permission
from thestage.controllers.utils_controller import validate_config_and_get_service_factory, get_current_directory
from thestage.helpers.logger.app_logger import app_logger
from thestage.i18n.translation import __
from thestage.inference_model.business.inference_model_service import InferenceModelService
from thestage.logging.business.logging_service import LoggingService
from thestage.project.business.project_service import ProjectService

app = typer.Typer(no_args_is_help=True, help="Manage project inference simulator models")
@app.command("ls", help=__("List inference simulator models"), **get_command_metadata(CliCommand.PROJECT_MODEL_LS))
def list_inference_simulator_models(
        project_public_id: Optional[str] = typer.Option(
            None,
            '--project-id',
            '-pid',
            help=__("Project ID. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        project_slug: Optional[str] = typer.Option(
            None,
            '--project-name',
            '-pn',
            help=__("Project name. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
        statuses: List[str] = typer.Option(
            None,
            '--status',
            '-s',
            help=__("Filter by status, use --status all to list all inference simulator models"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_MODEL_LS
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [project_public_id, project_slug]) > 1:
        typer.echo("Provide a single identifier for project - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    inference_model_service: InferenceModelService = service_factory.get_inference_model_service()

    inference_model_service.print_inference_simulator_model_list(
        project_public_id=project_public_id,
        project_slug=project_slug,
        statuses=statuses,
        row=row,
        page=page
    )

    typer.echo(__("Inference simulator models listing complete"))
    raise typer.Exit(0)


@app.command("deploy-instance", no_args_is_help=True, help=__("Deploy inference simulator model to an instance"), **get_command_metadata(CliCommand.PROJECT_MODEL_DEPLOY_INSTANCE))
def deploy_inference_simulator_model_to_instance(
        model_public_id: Optional[str] = typer.Option(
            None,
            '--model-id',
            '-mid',
            help="Inference simulator model ID",
            is_eager=False,
        ),
        model_slug: Optional[str] = typer.Option(
            None,
            '--model-name',
            '-mn',
            help="Inference simulator model name",
            is_eager=False,
        ),
        rented_instance_public_id: Optional[str] = typer.Option(
            None,
            '--rented-instance-id',
            '-rid',
            help=__("Rented instance ID on which the inference simulator will run"),
            is_eager=False,
        ),
        rented_instance_slug: Optional[str] = typer.Option(
            None,
            '--rented-instance-name',
            '-rn',
            help=__("Rented instance name on which the inference simulator will run"),
            is_eager=False,
        ),
        self_hosted_instance_public_id: Optional[str] = typer.Option(
            None,
            '--self-hosted-instance-id',
            '-sid',
            help=__("Self-hosted instance ID on which the inference simulator will run"),
            is_eager=False,
        ),
        self_hosted_instance_slug: Optional[str] = typer.Option(
            None,
            '--self-hosted-instance-name',
            '-sn',
            help=__("Self-hosted instance name on which the inference simulator will run"),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory. By default, the current directory is used"),
            show_default=False,
            is_eager=False,
        ),
        enable_log_stream: Optional[bool] = typer.Option(
            True,
            "--no-logs",
            "-nl",
            help=__("Disable real-time log streaming"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_MODEL_DEPLOY_INSTANCE
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if model_slug and not re.match(r"^[a-zA-Z0-9-]+$", model_slug):
        raise typer.BadParameter(__("Invalid UID format. UID can only contain letters, numbers, and hyphens."))

    new_inference_simulator_slug = f"{model_slug}-{int(time.time())}"

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    inference_model_service: InferenceModelService = service_factory.get_inference_model_service()

    inference_simulator_public_id=inference_model_service.project_deploy_inference_simulator_model_to_instance(
        model_public_id=model_public_id,
        model_slug=model_slug,
        new_inference_simulator_slug=new_inference_simulator_slug,
        rented_instance_public_id=rented_instance_public_id,
        rented_instance_slug=rented_instance_slug,
        self_hosted_instance_public_id=self_hosted_instance_public_id,
        self_hosted_instance_slug=self_hosted_instance_slug,
    )

    if enable_log_stream:
        logging_service: LoggingService = service_factory.get_logging_service()

        logging_service.stream_inference_simulator_logs_with_controls(
            public_id=inference_simulator_public_id
        )
    raise typer.Exit(0)


@app.command("deploy-sagemaker", no_args_is_help=True, help=__("Deploy inference simulator model to SageMaker"), **get_command_metadata(CliCommand.PROJECT_MODEL_DEPLOY_SAGEMAKER))
def deploy_inference_simulator_model_to_sagemaker(
        model_public_id: Optional[str] = typer.Option(
            None,
            '--model-id',
            '-mid',
            help="Inference simulator model ID",
            is_eager=False,
        ),
        model_slug: Optional[str] = typer.Option(
            None,
            '--model-name',
            '-mn',
            help="Inference simulator model name",
            is_eager=False,
        ),
        arn: Optional[str] = typer.Option(
            None,
            '--amazon-resource-name',
            '-arn',
            help=__("Amazon Resource Name of the IAM Role to use, e.g., arn:aws:iam::{aws_account_id}:role/{role}"),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory. By default, the current directory is used"),
            show_default=False,
            is_eager=False,
        ),
        instance_type: Optional[str] = typer.Option(
            None,
            '--instance-type',
            '-it',
            help=__("Instance type on which the inference simulator model will be deployed"),
            is_eager=False,
        ),
        initial_variant_weight: Optional[float] = typer.Option(
            None,
            "--initial-variant-weight",
            "-ivw",
            help=__("Initial Variant Weight. By default 1.0"),
            show_default=False,
            is_eager=False,
        ),
        initial_instance_count: Optional[int] = typer.Option(
            None,
            "--initial-instance-count",
            "-iic",
            help=__("Initial Instance Count"),
            show_default=False,
            is_eager=False,
        ),

):
    command_name = CliCommand.PROJECT_MODEL_DEPLOY_SAGEMAKER
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [model_public_id, model_slug]) != 1:
        typer.echo("Provide a single identifier for inference simulator model - ID or name.")
        raise typer.Exit(1)

    if model_slug and not re.match(r"^[a-zA-Z0-9-]+$", model_slug):
        raise typer.BadParameter(__("Invalid UID format. UID can only contain letters, numbers, and hyphens."))

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    inference_model_service: InferenceModelService = service_factory.get_inference_model_service()

    inference_model_service.project_deploy_inference_simulator_model_to_sagemaker(
        model_public_id=model_public_id,
        model_slug=model_slug,
        arn=arn,
        instance_type=instance_type,
        initial_variant_weight=initial_variant_weight,
        initial_instance_count=initial_instance_count,
    )



from typing import Optional, List

from pydantic import Field, BaseModel

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.services.clients.thestage_api.dtos.frontend_status import FrontendStatusDto


class EntityMatchData(BaseModel):
    publicId: str = Field(alias='publicId')
    frontendStatus: FrontendStatusDto = Field(alias='frontendStatus')
    canConnect: bool = Field(alias='canConnect')
    canDownloadUploadOnContainer: bool = Field(alias='canDownloadUploadOnContainer')
    matchedField: Optional[str] = Field(None, alias='matchedField')


class ConnectResolveOptionsResponse(TheStageBaseResponse):
    taskMatchData: List[EntityMatchData] = Field(None, alias='taskMatchData')
    dockerContainerMatchData: List[EntityMatchData] = Field(None, alias='dockerContainerMatchData')
    instanceRentedMatchData: List[EntityMatchData] = Field(None, alias='instanceRentedMatchData')
    selfhostedInstanceMatchData: List[EntityMatchData] = Field(None, alias='selfhostedInstanceMatchData')



from typing import Optional

from thestage.config.business.config_provider import ConfigProvider
from thestage.services.clients.thestage_api.core.api_client_core import TheStageApiClientCore
from thestage.connect.dto.add_ssh_key_to_user_request import AddSshKeyToUserRequest
from thestage.connect.dto.add_ssh_key_to_user_response import AddSshKeyToUserResponse
from thestage.connect.dto.is_user_has_public_ssh_key_request import IsUserHasSshPublicKeyRequest
from thestage.connect.dto.is_user_has_public_ssh_key_response import IsUserHasSshPublicKeyResponse
from thestage.connect.dto.add_ssh_public_key_to_instance_request import AddSshPublicKeyToInstanceRequest
from thestage.connect.dto.add_ssh_public_key_to_instance_response import AddSshPublicKeyToInstanceResponse
from thestage.connect.dto.connect_resolve_response import ConnectResolveOptionsResponse


class ConnectApiClient(TheStageApiClientCore):
    def __init__(self, config_provider: ConfigProvider):
        super().__init__(url=config_provider.get_config().main.thestage_api_url)
        self.__config_provider = config_provider

    def add_public_ssh_key_to_user(self, public_key: str, note: str) -> AddSshKeyToUserResponse:
        request = AddSshKeyToUserRequest(
            sshKey=public_key,
            note=note
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/ssh-key/add-public-key-to-user',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = AddSshKeyToUserResponse.model_validate(response) if response else None
        return result


    def is_user_has_ssh_public_key(self, public_key: str) -> IsUserHasSshPublicKeyResponse:
        request = IsUserHasSshPublicKeyRequest(
            sshKey=public_key,
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/ssh-key/is-user-has-public-ssh-key',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = IsUserHasSshPublicKeyResponse.model_validate(response) if response else None
        return result


    def add_public_ssh_key_to_instance_rented(self, instance_rented_public_id: str,
                                              ssh_key_pair_public_id: str) -> AddSshPublicKeyToInstanceResponse:
        request = AddSshPublicKeyToInstanceRequest(
            instanceRentedPublicId=instance_rented_public_id,
            sshPublicKeyPublicId=ssh_key_pair_public_id,
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/ssh-key/add-public-ssh-key-to-instance-rented',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = AddSshPublicKeyToInstanceResponse.model_validate(response) if response else None
        return result

    def resolve_user_input(
            self,
            entity_identifier: str
    ) -> Optional[ConnectResolveOptionsResponse]:
        data = {
            "entityIdentifier": entity_identifier,
        }

        response = self._request(
            method='POST',
            url='/user-api/v1/resolve-user-input',
            data=data,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        return ConnectResolveOptionsResponse.model_validate(response) if response else None


import typer

from thestage.cli_command import CliCommand
from thestage.cli_command_helper import get_command_metadata, check_command_permission
from thestage.controllers.utils_controller import validate_config_and_get_service_factory
from thestage.docker_container.business.container_service import ContainerService
from thestage.i18n.translation import __

app = typer.Typer(no_args_is_help=True, help=__("Manage Docker images"))

@app.command(name='ls', help=__("List available docker images"), **get_command_metadata(CliCommand.CONTAINER_LS))
def list_available_images(
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONTAINER_IMAGE_LS
    check_command_permission(command_name)

    service_factory = validate_config_and_get_service_factory()
    container_service: ContainerService = service_factory.get_container_service()

    container_service.print_docker_image_list(
        row=row,
        page=page,
    )

    typer.echo(__("Images listing complete"))
    raise typer.Exit(0)


from pydantic import Field, ConfigDict

from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.inference_simulator.dto.inference_simulator import InferenceSimulator


class StartInferenceSimulatorResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)

    inferenceSimulator: InferenceSimulator = Field(None, alias='inferenceSimulator')



from typing import Optional, Any, Dict
from pydantic import BaseModel, ConfigDict, Field

class InferenceModel(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    public_id: Optional[str] = Field(None, alias='publicId')
    slug: Optional[str] = Field(None, alias='slug')
    status: Optional[str] = Field(None, alias='status')
    environment_metadata: Optional[Dict[str, Any]] = Field(None, alias='environmentMetadata')
    commit_hash: Optional[str] = Field(None, alias='commitHash')
    created_at: Optional[str] = Field(None, alias='createdAt')
    updated_at: Optional[str] = Field(None, alias='updatedAt')



from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class ProjectEntity(BaseModel):
    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
    )

    public_id: Optional[str] = Field(None, alias='ID')
    slug: Optional[str] = Field(None, alias='NAME')
    status: Optional[str] = Field(None, alias='STATUS')
    last_commit_hash: Optional[str] = Field(None, alias='LAST COMMIT HASH')



from pathlib import Path
from typing import List, Optional

import click
import typer
from tabulate import tabulate

from thestage.instance.communication.instance_api_client import InstanceApiClient
from thestage.instance.dto.instance_rented_create_api_dtos import InstanceRentedCreateRequest, ConcreteInstanceOffer
from thestage.instance.dto.rented_instance import RentedInstanceEntity
from thestage.instance.dto.self_hosted_instance import SelfHostedInstanceEntity
from thestage.i18n.translation import __
from thestage.instance.dto.enum.selfhosted_status import SelfhostedBusinessStatus
from thestage.instance.dto.enum.instance_rented_status import InstanceRentedBusinessStatus
from thestage.services.abstract_service import AbstractService
from thestage.helpers.error_handler import error_handler
from thestage.instance.dto.instance_rented_response import InstanceRentedDto
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList
from thestage.instance.dto.selfhosted_instance_response import SelfHostedInstanceDto
from thestage.config.business.config_provider import ConfigProvider
from thestage.instance.business.mapper.rented_instance_mapper import RentedInstanceMapper
from thestage.instance.business.mapper.selfhosted_instance_mapper import SelfHostedInstanceMapper
from thestage.connect.business.remote_server_service import RemoteServerService


class InstanceService(AbstractService):
    __instance_api_client: InstanceApiClient = None
    __config_provider: ConfigProvider = None

    def __init__(
            self,
            instance_api_client: InstanceApiClient,
            config_provider: ConfigProvider,
            remote_server_service: RemoteServerService,
    ):
        self.__instance_api_client = instance_api_client
        self.__remote_server_service = remote_server_service
        self.__config_provider = config_provider


    @error_handler()
    def check_instance_status_to_connect(
            self,
            instance: InstanceRentedDto,
    ) -> InstanceRentedDto:
        if instance:
            if instance.frontend_status.status_key in [
                InstanceRentedBusinessStatus.IN_QUEUE.name,
                InstanceRentedBusinessStatus.CREATING.name,
                InstanceRentedBusinessStatus.REBOOTING.name,
                InstanceRentedBusinessStatus.STARTING.name,
            ]:
                typer.echo(__('Cannot connect to rented server instance: it is either in the process of being rented or rebooted'))
                raise typer.Exit(1)
            elif instance.frontend_status.status_key in [
                InstanceRentedBusinessStatus.TERMINATING.name,
                InstanceRentedBusinessStatus.RENTAL_ERROR.name,
            ]:
                typer.echo(__('Cannot connect to rented server instance: renting failed'))
                raise typer.Exit(1)
            elif instance.frontend_status.status_key in [
                InstanceRentedBusinessStatus.STOPPED.name,
                InstanceRentedBusinessStatus.STOPPING.name,
                InstanceRentedBusinessStatus.DELETED.name,
            ]:
                typer.echo(__('Cannot connect to rented server instance: it is either stopped or has been deleted'))
                raise typer.Exit(1)
            elif instance.frontend_status.status_key in [
                InstanceRentedBusinessStatus.UNKNOWN.name,
                InstanceRentedBusinessStatus.ALL.name,
            ]:
                typer.echo(__('Cannot connect to rented server instance: instance status unknown'))
                raise typer.Exit(1)

        return instance

    @error_handler()
    def check_selfhosted_status_to_connect(
            self,
            instance: SelfHostedInstanceDto,
    ) -> SelfHostedInstanceDto:
        if instance:
            if instance.frontend_status.status_key in [
                SelfhostedBusinessStatus.AWAITING_CONFIGURATION.name,
            ]:
                typer.echo(__('Cannot connect to self-hosted instance: it is awaiting configuration'))
                raise typer.Exit(1)
            elif instance.frontend_status.status_key in [
                SelfhostedBusinessStatus.UNREACHABLE_DAEMON.name,
                SelfhostedBusinessStatus.DELETED.name,
            ]:
                typer.echo(__('Cannot connect to self-hosted instance: it may be turned off or unreachable'))
                raise typer.Exit(1)
            elif instance.frontend_status.status_key in [
                SelfhostedBusinessStatus.UNKNOWN.name,
                SelfhostedBusinessStatus.ALL.name,
            ]:
                typer.echo(__('Cannot connect to self-hosted instance: instance status unknown'))
                raise typer.Exit(1)

        return instance

    @error_handler()
    def connect_to_rented_instance(
            self,
            instance_rented_public_id: Optional[str],
            instance_rented_slug: Optional[str],
            input_ssh_key_path: Optional[str]
    ):
        instance = self.__instance_api_client.get_rented_instance(
            instance_public_id=instance_rented_public_id,
            instance_slug=instance_rented_slug,
        )

        if instance:
            self.check_instance_status_to_connect(
                instance=instance,
            )

            ssh_path_from_config: Optional[str] = None
            if not input_ssh_key_path:
                ssh_path_from_config = self.__config_provider.get_valid_private_key_path_by_ip_address(instance.ip_address)
                if ssh_path_from_config:
                    typer.echo(f"Using configured SSH key for this instance: {ssh_path_from_config}")

            if not input_ssh_key_path and not ssh_path_from_config:
                typer.echo('Using SSH agent to connect to server instance')

            self.__remote_server_service.connect_to_instance(
                ip_address=instance.ip_address,
                username=instance.host_username,
                private_key_path=ssh_path_from_config or input_ssh_key_path
            )

            # cannot really detect how ssh connection was ended. capturing stderr using subprocess feels bad/unreliable.
            if input_ssh_key_path:
                self.__config_provider.update_remote_server_config_entry(ip_address=instance.ip_address, ssh_key_path=Path(input_ssh_key_path))
        else:
            typer.echo(__("Server instance not found: %instance_item%", {'instance_item': instance_rented_slug}))


    @error_handler()
    def connect_to_selfhosted_instance(
            self,
            selfhosted_instance_public_id: Optional[str],
            selfhosted_instance_slug: Optional[str],
            username: str,
            input_ssh_key_path: Optional[str],
    ):
        if not username:
            username = 'root'
            typer.echo(__("No remote server username provided, using 'root' as username"))

        instance = self.__instance_api_client.get_selfhosted_instance(
            instance_public_id=selfhosted_instance_public_id,
            instance_slug=selfhosted_instance_slug,
        )

        if instance:
            self.check_selfhosted_status_to_connect(
                instance=instance,
            )

            ssh_path_from_config: Optional[str] = None
            if not input_ssh_key_path:
                ssh_path_from_config = self.__config_provider.get_valid_private_key_path_by_ip_address(instance.ip_address)
                if ssh_path_from_config:
                    typer.echo(f"Using configured SSH key for this instance: {ssh_path_from_config}")

            if not input_ssh_key_path and not ssh_path_from_config:
                typer.echo('Using SSH agent to connect to server instance')

            self.__remote_server_service.connect_to_instance(
                ip_address=instance.ip_address,
                username=username,
                private_key_path=ssh_path_from_config or input_ssh_key_path
            )

            if input_ssh_key_path:
                self.__config_provider.update_remote_server_config_entry(ip_address=instance.ip_address, ssh_key_path=Path(input_ssh_key_path))
        else:
            typer.echo("Self-hosted instance not found")


    @error_handler()
    def get_rented_list(
            self,
            statuses: List[str],
            row: int = 5,
            page: int = 1,
    ) -> PaginatedEntityList[InstanceRentedDto]:
        data = self.__instance_api_client.get_rented_instance_list(
            statuses=statuses,
            page=page,
            limit=row,
        )

        return data

    @error_handler()
    def get_self_hosted_list(
            self,
            statuses: List[str],
            row: int = 5,
            page: int = 1,
    ) -> PaginatedEntityList[SelfHostedInstanceDto]:
        data = self.__instance_api_client.get_selfhosted_instance_list(
            statuses=statuses,
            page=page,
            limit=row,
        )
        return data


    @error_handler()
    def print_self_hosted_instance_list(self, statuses, row, page):
        selfhosted_instance_status_map = self.__instance_api_client.get_selfhosted_business_status_map()

        if not statuses:
            statuses = ({key: selfhosted_instance_status_map[key] for key in [
                SelfhostedBusinessStatus.AWAITING_CONFIGURATION,
                SelfhostedBusinessStatus.RUNNING,
                SelfhostedBusinessStatus.UNREACHABLE_DAEMON,
            ]}).values()

        if "all" in statuses:
            statuses = selfhosted_instance_status_map.values()

        for input_status_item in statuses:
            if input_status_item not in selfhosted_instance_status_map.values():
                typer.echo(__("'%invalid_status%' is not one of %valid_statuses%", {
                    'invalid_status': input_status_item,
                    'valid_statuses': str(list(selfhosted_instance_status_map.values()))
                }))
                raise typer.Exit(1)

        typer.echo(__(
            "Listing self-hosted instances with the following statuses: %statuses%, to view all self-hosted instances, use --status all",
            placeholders={
                'statuses': ', '.join([status_item for status_item in statuses])
            }))

        backend_statuses: List[str] = [key for key, value in selfhosted_instance_status_map.items() if value in statuses]

        self.print(
            func_get_data=self.get_self_hosted_list,
            func_special_params={
                'statuses': backend_statuses,
            },
            mapper=SelfHostedInstanceMapper(),
            headers=list(map(lambda x: x.alias, SelfHostedInstanceEntity.model_fields.values())),
            row=row,
            page=page,
            show_index="never",
        )


    @error_handler()
    def print_rented_instance_list(self, statuses, row, page):
        instance_rented_status_map = self.__instance_api_client.get_rented_business_status_map()

        if not statuses:
            statuses = ({key: instance_rented_status_map[key] for key in [
                InstanceRentedBusinessStatus.ONLINE,
                InstanceRentedBusinessStatus.CREATING,
                InstanceRentedBusinessStatus.TERMINATING,
                InstanceRentedBusinessStatus.REBOOTING,
            ]}).values()

        if "all" in statuses:
            statuses = instance_rented_status_map.values()

        for input_status_item in statuses:
            if input_status_item not in instance_rented_status_map.values():
                typer.echo(__("'%invalid_status%' is not one of %valid_statuses%", {
                    'invalid_status': input_status_item,
                    'valid_statuses': str(list(instance_rented_status_map.values()))
                }))
                raise typer.Exit(1)

        typer.echo(__(
            "Listing rented server instances with the following statuses: %statuses%, to view all rented server instances, use --status all",
            placeholders={
                'statuses': ', '.join([status_item for status_item in statuses])
            }))

        backend_statuses: List[str] = [key for key, value in instance_rented_status_map.items() if value in statuses]

        self.print(
            func_get_data=self.get_rented_list,
            func_special_params={
                'statuses': backend_statuses,
            },
            mapper=RentedInstanceMapper(),
            headers=list(map(lambda x: x.alias, RentedInstanceEntity.model_fields.values())),
            row=row,
            page=page,
            show_index="never",
        )


    @error_handler()
    def rent_instance(
            self,
            slug: str,
            gpu_name: str,
            gpu_count: Optional[int],
            cloud_provider_name: str,
            disk_size_gb: Optional[int],
            force_start_rent: bool
    ):
        request = InstanceRentedCreateRequest(
            slug=slug,
            gpuName=gpu_name,
            cloudProviderName=cloud_provider_name,
            gpuCount=None if gpu_count is None else str(gpu_count),
            diskSizeGb=None if disk_size_gb is None else str(disk_size_gb),
            forceStartRent=force_start_rent,
        )
        config = self.__config_provider.get_config()

        response = None

        if not force_start_rent:
            is_optional_arguments_settled = force_start_rent
            while not is_optional_arguments_settled:
                response = self.__instance_api_client.create_instance_rented(request=request)
                if not response.isSuccess:
                    if response.message:
                        typer.echo(response.message)
                    if 'gpuName' in response.fieldsWithError:
                        if config.runtime.command_metadata.instanceOfferGpuNames:
                            typer.echo(f"Invalid GPU name. Valid GPU names: {', '.join(config.runtime.command_metadata.instanceOfferGpuNames)}")
                        else:
                            typer.echo('Invalid GPU name')
                    if 'providerName' in response.fieldsWithError:
                        if config.runtime.command_metadata.instanceOfferProviderNames:
                            typer.echo(f"Invalid provider name. Valid provider names: {', '.join(config.runtime.command_metadata.instanceOfferProviderNames)}")
                        else:
                            typer.echo('Invalid provider name')
                    typer.Exit(1)

                if response.missingFieldName:
                    if not len(response.missingFieldOptions):
                        typer.echo("No rented instance offers found with the requested parameters")
                        raise typer.Exit(1)

                    prompt_result: str = typer.prompt(
                        text=f'Selection is required for: {response.missingFieldDisplayName}',
                        show_choices=True,
                        type=click.Choice([v for v in response.missingFieldOptions]),
                    )

                    # field type must be str to avoid type cast errors
                    setattr(request, response.missingFieldName, prompt_result)
                else:
                    is_optional_arguments_settled = True

            if not response.concreteInstanceOffersToSelect:
                # no offers found for requested params
                raise typer.Exit(1)

            self.print_instance_offers(concrete_offers=response.concreteInstanceOffersToSelect, show_public_id=False, show_index=True)

            offer_indices = range(1, len(response.concreteInstanceOffersToSelect) + 1)
            prompt_result: int = typer.prompt(
                text='Selection is required for: Offer #',
                show_choices=False,
                type=click.Choice([str(i) for i in offer_indices]),
            )

            selected_offer_public_id = response.concreteInstanceOffersToSelect[int(prompt_result) - 1].publicId
            request.concreteInstanceOfferPublicId = selected_offer_public_id

        response = self.__instance_api_client.create_instance_rented(request=request)

        if not response.isSuccess:
            typer.echo(response.message)
            raise typer.Exit(1)

        typer.echo('Using the following instance offer to rent the instance:')
        self.print_instance_offers(concrete_offers=[response.selectedConcreteInstanceOffer], show_public_id=True, show_index=False)
        typer.echo(f'Instance rental process initiated successfully. New instance rented ID: {response.instanceRentedPublicId}')


    def print_instance_offers(self, concrete_offers: List[ConcreteInstanceOffer], show_public_id: bool, show_index: bool):
        offer_table_data = [list(item.model_dump(exclude=({} if show_public_id else {'publicId'})).values()) for item in concrete_offers]
        headers = []
        for name, field in ConcreteInstanceOffer.model_fields.items():
            if show_public_id or name != 'publicId':
                headers.append(field.title)

        offer_indices = range(1, len(concrete_offers) + 1)

        typer.echo(tabulate(
            offer_table_data,
            headers=(['Offer #'] if show_index else []) + headers,
            showindex=offer_indices if show_index else False,
            tablefmt="double_grid",
        ))


    @error_handler()
    def terminate_instance_rented(
            self,
            instance_rented_public_id: Optional[str],
            instance_rented_slug: Optional[str]
    ):

        result = self.__instance_api_client.terminate_instance_rented(
            instance_rented_public_id=instance_rented_public_id,
            instance_rented_slug=instance_rented_slug,
        )

        for terminated_pubic_id_item in result.terminatedInstanceRentedPublicIds:
            typer.echo(f"Termination for rented instance '{terminated_pubic_id_item}' initiated successfully")

        if result.message:
            typer.echo(result.message)


    @error_handler()
    def restart_instance_rented(
            self,
            instance_rented_public_id: Optional[str],
            instance_rented_slug: Optional[str]
    ):

        result = self.__instance_api_client.restart_instance_rented(
            instance_rented_public_id=instance_rented_public_id,
            instance_rented_slug=instance_rented_slug,
        )

        if result.message:
            typer.echo(result.message)

        if not result.isSuccess:
            raise typer.Exit(1)



from typing import Optional, List, Dict

from thestage.config.business.config_provider import ConfigProvider
from thestage.global_dto.enums.order_direction_type import OrderDirectionType
from thestage.services.clients.thestage_api.core.api_client_core import TheStageApiClientCore
from thestage.services.clients.thestage_api.dtos.entity_filter_request import EntityFilterRequest
from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList
from thestage.task.dto.list_for_project_request import TaskListForProjectRequest
from thestage.task.dto.list_for_project_response import TaskListForProjectResponse
from thestage.task.dto.run_task_request import RunTaskRequest
from thestage.task.dto.run_task_response import RunTaskResponse
from thestage.task.dto.status_localized_map_response import TaskStatusLocalizedMapResponse
from thestage.task.dto.view_response import TaskViewResponse
from thestage.task.dto.task import Task


class TaskApiClient(TheStageApiClientCore):
    def __init__(self, config_provider: ConfigProvider):
        super().__init__(url=config_provider.get_config().main.thestage_api_url)
        self.__config_provider = config_provider

    def get_task(
            self,
            task_public_id: str,
    ) -> Optional[TaskViewResponse]:
        data = {
            "taskPublicId": task_public_id,
        }

        response = self._request(
            method='POST',
            url='/user-api/v2/task/view',
            data=data,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = TaskViewResponse.model_validate(response) if response else None
        return result if result and result.isSuccess else None

    def get_task_list_for_project(
            self,
            project_public_id: Optional[str],
            project_slug: Optional[str],
            page: int = 1,
            limit: int = 10,
    ) -> Optional[PaginatedEntityList[Task]]:
        request = TaskListForProjectRequest(
            projectPublicId=project_public_id,
            projectSlug=project_slug,
            entityFilterRequest=EntityFilterRequest(
                orderByField="createdAt",
                orderByDirection=OrderDirectionType.DESC,
                page=page,
                limit=limit,
            ),
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/task/list',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = TaskListForProjectResponse.model_validate(response) if response else None
        return result.tasks if result and result.isSuccess else None

    def get_task_localized_status_map(self) -> Optional[Dict[str, str]]:
        response = self._request(
            method='POST',
            url='/user-api/v1/task/status-localized-mapping',
            data=None,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        data = TaskStatusLocalizedMapResponse.model_validate(response) if response else None

        return data.taskStatusMap if data else None

    def cancel_task(
            self,
            task_public_id: str,
    ) -> Optional[TheStageBaseResponse]:
        data = {
            "taskPublicId": task_public_id,
        }

        response = self._request(
            method='POST',
            url='/user-api/v2/task/cancel',
            data=data,
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        result = TheStageBaseResponse.model_validate(response) if response else None
        return result if result else None

    def execute_project_task(
            self,
            project_public_id: str,
            run_command: str,
            task_title: str,
            docker_container_public_id: str,
            commit_hash: Optional[str] = None,
    ) -> Optional[RunTaskResponse]:
        request = RunTaskRequest(
            projectPublicId=project_public_id,
            dockerContainerPublicId=docker_container_public_id,
            commitHash=commit_hash,
            runCommand=run_command,
            taskTitle=task_title,
        )

        response = self._request(
            method='POST',
            url='/user-api/v2/task/execute',
            data=request.model_dump(),
            token=self.__config_provider.get_config().main.thestage_auth_token,
        )

        return RunTaskResponse.model_validate(response) if response else None



from typing import Optional, Any, Dict
from pydantic import BaseModel, ConfigDict, Field

class InferenceSimulator(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    public_id: Optional[str] = Field(None, alias="publicId")
    slug: Optional[str] = Field(None, alias="slug")
    status: Optional[str] = Field(None, alias="status")
    http_endpoint: Optional[str] = Field(None, alias="httpEndpoint")
    grpc_endpoint: Optional[str] = Field(None, alias="grpcEndpoint")
    metrics_endpoint: Optional[str] = Field(None, alias="metricsEndpoint")
    qlip_serve_metadata: Optional[Dict[str, Any]] = Field(None, alias="qlipServeMetadata")
    created_at: Optional[str] = Field(None, alias="createdAt")



from pydantic import Field, ConfigDict, BaseModel


class AddSshKeyToUserRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    sshKey: str = Field(None, alias='sshKey')
    note: str = Field(None, alias='note')



from typing import Optional

from thestage.instance.dto.self_hosted_instance import SelfHostedInstanceEntity
from thestage.instance.dto.enum.gpu_name import InstanceGpuType
from thestage.instance.dto.selfhosted_instance_response import SelfHostedInstanceDto
from thestage.services.abstract_mapper import AbstractMapper


class SelfHostedInstanceMapper(AbstractMapper):

    def build_entity(self, item: SelfHostedInstanceDto) -> Optional[SelfHostedInstanceEntity]:
        if not item:
            return None

        gpus = []
        if item.detected_gpus:
            gpus = [item.type.value for item in item.detected_gpus.gpus]

        if len(gpus) == 0:
            gpus = [InstanceGpuType.NO_GPU]

        return SelfHostedInstanceEntity(
            slug=item.slug,
            public_id=item.public_id,
            cpu_type=item.cpu_type,
            cpu_cores=item.cpu_cores,
            gpu_type=', '.join(gpus),
            ip_address=item.ip_address,
            status=item.frontend_status.status_translation if item.frontend_status else None,
        )



from typing import List, Optional

from pydantic import Field, ConfigDict, BaseModel


class DeleteProjectRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    projectPublicId: Optional[str] = Field(None, alias='projectPublicId')
    projectSlug: Optional[str] = Field(None, alias='projectSlug')


from typing import List, Optional

from pydantic import Field, ConfigDict, BaseModel

from thestage.services.clients.thestage_api.dtos.entity_filter_request import EntityFilterRequest


class ProjectListRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    entityFilterRequest: EntityFilterRequest = Field(None, alias='entityFilterRequest')
    statuses: Optional[List[str]] = Field(None, alias='statuses')
    rentedInstancePublicId: Optional[str] = Field(None, alias='rentedInstancePublicId')
    rentedInstanceSlug: Optional[str] = Field(None, alias='rentedInstanceSlug')
    selfHostedInstancePublicId: Optional[str] = Field(None, alias='selfHostedInstancePublicId')
    selfHostedInstanceSlug: Optional[str] = Field(None, alias='selfHostedInstanceSlug')


from typing import Dict

from pydantic import ConfigDict, Field

from thestage.project.dto.project_response import ProjectDto
from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList


class ProjectListResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)

    paginatedList: PaginatedEntityList[ProjectDto] = Field(None, alias='paginatedList')

class ProjectBusinessStatusMapperResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)
    project_business_status_map: Dict[str, str] = Field(default={}, alias='projectStatusMap')


from typing import Optional
import typer

from thestage.cli_command import CliCommand
from thestage.cli_command_helper import check_command_permission
from thestage.color_scheme.color_scheme import ColorScheme
from thestage.connect.communication.connect_api_client import ConnectApiClient
from thestage.i18n.translation import __
from thestage.instance.communication.instance_api_client import InstanceApiClient
from thestage.services.clients.thestage_api.core.http_client_exception import HttpClientException
from thestage.instance.dto.enum.instance_rented_status import InstanceRentedBusinessStatus
from thestage.services.abstract_service import AbstractService
from thestage.helpers.error_handler import error_handler
from thestage.instance.dto.instance_rented_response import InstanceRentedDto
from thestage.docker_container.business.container_service import ContainerService
from thestage.instance.business.instance_service import InstanceService
from thestage.logging.business.logging_service import LoggingService
from rich import print


class ConnectService(AbstractService):
    __connect_api_client: ConnectApiClient = None
    __instance_api_client: InstanceApiClient = None
    __instance_service: InstanceService = None
    __container_service: ContainerService = None
    __logging_service: LoggingService = None

    def __init__(
            self,
            instance_api_client: InstanceApiClient,
            connect_api_client: ConnectApiClient,
            instance_service: InstanceService,
            container_service: ContainerService,
            logging_service: LoggingService,
    ):
        super(ConnectService, self).__init__(
        )
        self.__instance_api_client = instance_api_client
        self.__connect_api_client = connect_api_client
        self.__instance_service = instance_service
        self.__container_service = container_service
        self.__logging_service = logging_service


    @error_handler()
    def connect_to_entity(
            self,
            input_entity_identifier: str,
            username: Optional[str],
            private_key_path: Optional[str],
    ):
        resolved_options = self.__connect_api_client.resolve_user_input(entity_identifier=input_entity_identifier)
        entities_available_for_connect_count = 0
        task_presence = False
        container_presence = False
        rented_presence = False
        selfhosted_presence = False
        resolved_entity_public_id = None

        if resolved_options.taskMatchData:
            for task_item in resolved_options.taskMatchData:
                message = f"Found a task with matching {task_item.matchedField} in status: '{task_item.frontendStatus.status_translation}' (ID: {task_item.publicId})"
                line_color = ColorScheme.SUCCESS if task_item.canConnect else 'default'
                print(f"[{line_color}]{message}[{line_color}]")
                if task_item.canConnect:
                    task_presence = True
                    resolved_entity_public_id = task_item.publicId
                    entities_available_for_connect_count += 1

        if resolved_options.dockerContainerMatchData:
            for container_item in resolved_options.dockerContainerMatchData:
                message = f"Found a container with matching {container_item.matchedField} in status: '{container_item.frontendStatus.status_translation}' (ID: {container_item.publicId})"
                line_color = ColorScheme.SUCCESS.value if container_item.canConnect else 'default'
                print(f"[{line_color}]{message}[{line_color}]")
                if container_item.canConnect:
                    container_presence = True
                    resolved_entity_public_id = container_item.publicId
                    entities_available_for_connect_count += 1

        if resolved_options.instanceRentedMatchData:
            for instance_rented_item in resolved_options.instanceRentedMatchData:
                message = f"Found a rented instance with matching {instance_rented_item.matchedField} in status: '{instance_rented_item.frontendStatus.status_translation}'  (ID: {instance_rented_item.publicId})"
                line_color = ColorScheme.SUCCESS.value if instance_rented_item.canConnect else 'default'
                print(f"[{line_color}]{message}[{line_color}]")

                if instance_rented_item.canConnect:
                    rented_presence = True
                    resolved_entity_public_id = instance_rented_item.publicId
                    entities_available_for_connect_count += 1

        if resolved_options.selfhostedInstanceMatchData:
            for selfhosted_item in resolved_options.selfhostedInstanceMatchData:
                message = f"Found a self-hosted instance with matching {selfhosted_item.matchedField} in status: '{selfhosted_item.frontendStatus.status_translation}' (ID: {selfhosted_item.publicId})"
                line_color = ColorScheme.SUCCESS.value if selfhosted_item.canConnect else 'default'
                print(f"[{line_color}]{message}[{line_color}]")

                if selfhosted_item.canConnect:
                    selfhosted_presence = True
                    resolved_entity_public_id = selfhosted_item.publicId
                    entities_available_for_connect_count += 1

        if entities_available_for_connect_count > 1:
            typer.echo("Provided identifier caused ambiguity")
            typer.echo("Consider running a dedicated command to connect to the entity you need")
            raise typer.Exit(code=1)

        if entities_available_for_connect_count == 0:
            typer.echo("There is nothing to connect to with the provided identifier")
            raise typer.Exit(code=1)

        if rented_presence:
            check_command_permission(CliCommand.INSTANCE_RENTED_CONNECT)
            typer.echo(f"Connecting to rented instance '{resolved_entity_public_id}'...")
            self.__instance_service.connect_to_rented_instance(
                instance_rented_public_id=resolved_entity_public_id,
                instance_rented_slug=None,
                input_ssh_key_path=private_key_path
            )

        if container_presence:
            check_command_permission(CliCommand.CONTAINER_CONNECT)
            typer.echo(f"Connecting to docker container '{resolved_entity_public_id}'...")
            self.__container_service.connect_to_container(
                container_public_id=resolved_entity_public_id,
                container_slug=None,
                username=username,
                input_ssh_key_path=private_key_path
            )

        if selfhosted_presence:
            check_command_permission(CliCommand.INSTANCE_SELF_HOSTED_CONNECT)
            typer.echo(f"Connecting to self-hosted instance '{resolved_entity_public_id}'...")

            self.__instance_service.connect_to_selfhosted_instance(
                selfhosted_instance_public_id=resolved_entity_public_id,
                selfhosted_instance_slug=None,
                username=username,
                input_ssh_key_path=private_key_path
            )

        if task_presence:
            typer.echo(f"Connecting to task '{resolved_entity_public_id}'...")
            self.__logging_service.stream_task_logs_with_controls(task_public_id=resolved_entity_public_id)


    @error_handler()
    def upload_ssh_key(self, public_key_contents: str, instance_public_id: Optional[str], instance_slug: Optional[str]):
        instance_rented: Optional[InstanceRentedDto] = None
        if instance_slug or instance_public_id:
            try:
                instance_rented = self.__instance_api_client.get_rented_instance(
                    instance_public_id=instance_public_id,
                    instance_slug=instance_slug
                )
            except HttpClientException as e:
                instance_rented = None

            if instance_rented is None:
                typer.echo(f"No rented instance found with matching identifier")
                raise typer.Exit(1)

        note_to_send: Optional[str] = None

        is_user_already_has_key_response = self.__connect_api_client.is_user_has_ssh_public_key(
            public_key=public_key_contents
        )

        ssh_key_pair_public_id = is_user_already_has_key_response.sshKeyPairPublicId
        is_adding_key_to_user = not is_user_already_has_key_response.isUserHasPublicKey

        if is_adding_key_to_user and not note_to_send:
            note_to_send: str = typer.prompt(
                text=__('SSH key will be added to your profile. Provide a name for this key'),
                show_choices=False,
                type=str,
                show_default=False,
            )

        if not is_adding_key_to_user and not instance_rented:
            typer.echo("Key already exists in your profile")

        if is_adding_key_to_user:
            add_ssh_key_to_user_response = self.__connect_api_client.add_public_ssh_key_to_user(
                public_key=public_key_contents,
                note=note_to_send
            )
            typer.echo(f"Public key '{note_to_send}' added to your profile")
            ssh_key_pair_public_id = add_ssh_key_to_user_response.sshKeyPairPublicId

        if instance_rented:
            self.__connect_api_client.add_public_ssh_key_to_instance_rented(
                instance_rented_public_id=instance_rented.public_id,
                ssh_key_pair_public_id=ssh_key_pair_public_id
            )

            if instance_rented.frontend_status.status_key != InstanceRentedBusinessStatus.ONLINE:
                typer.echo(f"Rented instance '{instance_rented.slug}' status is '{instance_rented.frontend_status.status_translation}'. Key will be added as soon as it is back online.")
            else:
                typer.echo(f"Public key added to rented instance '{instance_rented.slug}'")



from typing import Optional, List, Dict

from pydantic import Field, BaseModel, ConfigDict

from thestage.cli_command import CliCommandMetadata
from thestage.services.clients.thestage_api.dtos.base_response import TheStageBaseResponse, TheStageBasePaginatedResponse


class ValidateTokenResponse(TheStageBaseResponse):
    model_config = ConfigDict(use_enum_values=True)

    allowedCliCommands: Optional[Dict[str, bool]] = Field(default={}, alias='allowedCliCommands')
    cliCommandMetadata: Optional[CliCommandMetadata] = Field(default=None, alias='cliCommandMetadata')



from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class TaskEntity(BaseModel):

    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
    )

    public_id: Optional[str] = Field(None, alias='ID')
    title: Optional[str] = Field(None, alias='TITLE')
    status: Optional[str] = Field(None, alias='STATUS')
    docker_container_public_id: Optional[str] = Field(None, alias='CONTAINER ID')
    started_at: Optional[str] = Field(None, alias='STARTED AT')
    finished_at: Optional[str] = Field(None, alias='FINISHED_AT')



from typing import Optional

from pydantic import Field, ConfigDict, BaseModel


class LogPollingRequest(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    taskPublicId: Optional[str] = Field(None, alias='taskPublicId')
    inferenceSimulatorPublicId: Optional[str] = Field(None, alias='inferenceSimulatorPublicId')
    lastLogId: Optional[str] = Field(None, alias='lastLogId')
    dockerContainerPublicId: Optional[str] = Field(None, alias='dockerContainerPublicId')
    lastLogTimestamp: Optional[str] = Field(None, alias='lastLogTimestamp')



# Init project settings by ENV environment variable
import locale
from pathlib import Path


from python_gettext_translations.translations import init_translations

from thestage.config.env_base import *


THESTAGE_LOCAL_LANGUAGE = 'en_GB'
if locale.getlocale():
    THESTAGE_LOCAL_LANGUAGE = locale.getlocale()[0]

translation = Path(f'i18n/')

if translation.exists() and translation.is_dir():
    init_translations(f'i18n/')


from typing import Optional, Dict, List

from pydantic import BaseModel, ConfigDict, Field


class DockerContainerInstalledServiceDto(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    name: Optional[str] = Field(None, alias='name')
    port: Optional[int] = Field(None, alias='port')
    url: Optional[str] = Field(None, alias='url')
    variables: Dict[str, str] = Field(default_factory=dict, alias='variables')


class DockerContainerInstalledServicesDto(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    installed_services: List[DockerContainerInstalledServiceDto] = Field(default_factory=list, alias='installedServices')


from pathlib import Path
from typing import Optional, List

from thestage.cli_command import CliCommand
from thestage.cli_command_helper import get_command_group_help_panel, get_command_metadata, check_command_permission
from thestage.config.business import config_storage
from thestage.helpers.logger.app_logger import app_logger
from thestage.instance.business.instance_service import InstanceService
from thestage.i18n.translation import __
from thestage.controllers.utils_controller import \
    validate_config_and_get_service_factory, get_current_directory

import typer


app = typer.Typer(no_args_is_help=True, help=__("Manage server instances"))

rented = typer.Typer(no_args_is_help=True, help=__("Manage rented server instances"))
self_hosted = typer.Typer(no_args_is_help=True, help=__("Manage self-hosted instances"))

app.add_typer(rented, name="rented", rich_help_panel=get_command_group_help_panel())
app.add_typer(self_hosted, name="self-hosted", rich_help_panel=get_command_group_help_panel())


@rented.command(name="ls", help=__("List rented server instances"), **get_command_metadata(CliCommand.INSTANCE_RENTED_LS))
def rented_list(
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
        statuses: List[str] = typer.Option(
            None,
            '--status',
            '-s',
            help=__("Filter by status, use --status all to list all rented server instances"),
            is_eager=False,
        ),
):
    command_name = CliCommand.INSTANCE_RENTED_LS
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    service_factory = validate_config_and_get_service_factory()
    instance_service: InstanceService = service_factory.get_instance_service()

    instance_service.print_rented_instance_list(
        statuses=statuses,
        row=row,
        page=page
    )

    typer.echo(__("Rented server instances listing complete"))
    raise typer.Exit(0)


@rented.command(name="connect", no_args_is_help=True, help=__("Connect to rented server instance"), **get_command_metadata(CliCommand.INSTANCE_RENTED_CONNECT))
def instance_connect(
        public_id: Optional[str] = typer.Option(
            None,
            '--rented-instance-id',
            '-rid',
            help="Rented instance ID",
            is_eager=False,
        ),
        slug: Optional[str] = typer.Option(
            None,
            '--rented-instance-name',
            '-rn',
            help="Rented instance name",
            is_eager=False,
        ),
        private_ssh_key_path: str = typer.Option(
            None,
            "--private-key-path",
            "-pk",
            help=__("Path to private key that will be accepted by remote server (optional)"),
            is_eager=False,
        )
):
    command_name = CliCommand.INSTANCE_RENTED_CONNECT
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [public_id, slug]) != 1:
        typer.echo("Provide a single identifier for rented instance - ID or name.")
        raise typer.Exit(1)

    if private_ssh_key_path and not Path(private_ssh_key_path).is_file():
        typer.echo(f'No file found at provided path {private_ssh_key_path}')
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    instance_service: InstanceService = service_factory.get_instance_service()

    instance_service.connect_to_rented_instance(
        instance_rented_public_id=public_id,
        instance_rented_slug=slug,
        input_ssh_key_path=private_ssh_key_path
    )

    app_logger.info(f'Connection to rented instance closed')
    raise typer.Exit(0)


@self_hosted.command(name="ls", help=__("List self-hosted instances"), **get_command_metadata(CliCommand.INSTANCE_SELF_HOSTED_LS))
def self_hosted_list(
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
        statuses: List[str] = typer.Option(
            None,
            '--status',
            '-s',
            help=__("Filter by status, use --status all to list all self-hosted server instances"),
            is_eager=False,
        ),
):
    command_name = CliCommand.INSTANCE_SELF_HOSTED_LS
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    service_factory = validate_config_and_get_service_factory()
    instance_service: InstanceService = service_factory.get_instance_service()

    instance_service.print_self_hosted_instance_list(
        statuses=statuses,
        row=row,
        page=page
    )

    typer.echo(__("Self-hosted instances listing complete"))
    raise typer.Exit(0)


@self_hosted.command(name="connect", no_args_is_help=True, help=__("Connect to self-hosted instance"), **get_command_metadata(CliCommand.INSTANCE_SELF_HOSTED_CONNECT))
def self_hosted_connect(
        public_id: Optional[str] = typer.Option(
            None,
            '--self-hosted-instance-id',
            '-sid',
            help="Self-hosted instance ID",
            is_eager=False,
        ),
        slug: Optional[str] = typer.Option(
            None,
            '--self-hosted-instance-name',
            '-sn',
            help="Self-hosted instance name",
            is_eager=False,
        ),
        username: Optional[str] = typer.Option(
            None,
            '--username',
            '-u',
            help=__("Username for server instance (required when connecting to self-hosted instance)"),
            is_eager=False,
        ),
        private_ssh_key_path: str = typer.Option(
            None,
            "--private-key-path",
            "-pk",
            help=__("Path to private key that will be accepted by remote server (optional)"),
            is_eager=False,
        ),
):
    command_name = CliCommand.INSTANCE_SELF_HOSTED_CONNECT
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [public_id, slug]) != 1:
        typer.echo("Provide a single identifier for self-hosted instance - ID or name.")
        raise typer.Exit(1)

    if private_ssh_key_path and not Path(private_ssh_key_path).is_file():
        typer.echo(f'No file found at provided path {private_ssh_key_path}')
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    instance_service: InstanceService = service_factory.get_instance_service()

    instance_service.connect_to_selfhosted_instance(
        selfhosted_instance_public_id=public_id,
        selfhosted_instance_slug=slug,
        username=username,
        input_ssh_key_path=private_ssh_key_path
    )

    app_logger.info(f'Connection to self-hosted instance closed')
    raise typer.Exit(0)



@rented.command(name="create", no_args_is_help=True, help="Rent a server instance", **get_command_metadata(CliCommand.INSTANCE_RENTED_CREATE))
def instance_rented_create(
        slug: Optional[str] = typer.Option(
            None,
            '--name',
            '-n',
            help="New rented instance name",
            is_eager=False,
        ),
        cloud_provider_name: str = typer.Option(
            None,
            "--provider",
            "-p",
            help="Provider name" + (f" (allowed options: {str.join(', ', config_storage.APP_CONFIG.runtime.command_metadata.instanceOfferProviderNames)})" if config_storage.APP_CONFIG.runtime.command_metadata is not None and config_storage.APP_CONFIG.runtime.command_metadata.instanceOfferProviderNames else ''),
            is_eager=False,
        ),
        gpu_name: str = typer.Option(
            None,
            "--gpu",
            "-g",
            help="GPU name" + (f" (allowed options: {str.join(', ', config_storage.APP_CONFIG.runtime.command_metadata.instanceOfferGpuNames)})" if config_storage.APP_CONFIG.runtime.command_metadata is not None and config_storage.APP_CONFIG.runtime.command_metadata.instanceOfferGpuNames else ''),
            is_eager=False,
        ),
        gpu_count: int = typer.Option(
            None,
            "--num-gpu",
            "-ng",
            help="Number of GPUs",
            is_eager=False,
        ),
        disk_size_gb: int = typer.Option(
            None,
            "--disk-size",
            "-ds",
            help="Disk size in GB",
            is_eager=False,
        ),
        force_start_rent: bool = typer.Option(
            False,
            "--force-yes",
            "-f",
            help="Initiate rental process without additional prompts",
            is_eager=False,
            is_flag=True,
        ),

):
    command_name = CliCommand.INSTANCE_RENTED_CREATE
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if slug is None:
        typer.echo("Name (-n) is required")
        raise typer.Exit(1)

    if gpu_name is None:
        typer.echo("GPU name (-g) is required")
        raise typer.Exit(1)

    if cloud_provider_name is None:
        typer.echo("Provider name (-p) is required")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    instance_service: InstanceService = service_factory.get_instance_service()

    instance_service.rent_instance(
        slug=slug,
        gpu_name=gpu_name,
        gpu_count=gpu_count,
        cloud_provider_name=cloud_provider_name,
        disk_size_gb=disk_size_gb,
        force_start_rent=force_start_rent
    )

    raise typer.Exit(0)


@rented.command(name="terminate", no_args_is_help=True, help="Terminate a rented instance", **get_command_metadata(CliCommand.INSTANCE_RENTED_TERMINATE))
def instance_connect(
        public_id: Optional[str] = typer.Option(
            None,
            '--rented-instance-id',
            '-rid',
            help="Rented instance ID",
            is_eager=False,
        ),
        slug: Optional[str] = typer.Option(
            None,
            '--rented-instance-name',
            '-rn',
            help="Rented instance name",
            is_eager=False,
        ),
):
    command_name = CliCommand.INSTANCE_RENTED_TERMINATE
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [public_id, slug]) != 1:
        typer.echo("Provide a single identifier for rented instance - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    instance_service: InstanceService = service_factory.get_instance_service()

    instance_service.terminate_instance_rented(
        instance_rented_public_id=public_id,
        instance_rented_slug=slug
    )

    raise typer.Exit(0)


@rented.command(name="restart", no_args_is_help=True, help="Restart a rented instance", **get_command_metadata(CliCommand.INSTANCE_RENTED_RESTART))
def instance_connect(
        public_id: Optional[str] = typer.Option(
            None,
            '--rented-instance-id',
            '-rid',
            help="Rented instance ID",
            is_eager=False,
        ),
        slug: Optional[str] = typer.Option(
            None,
            '--rented-instance-name',
            '-rn',
            help="Rented instance name",
            is_eager=False,
        ),
):
    command_name = CliCommand.INSTANCE_RENTED_RESTART
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [public_id, slug]) != 1:
        typer.echo("Provide a single identifier for rented instance - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    instance_service: InstanceService = service_factory.get_instance_service()

    instance_service.restart_instance_rented(
        instance_rented_public_id=public_id,
        instance_rented_slug=slug
    )

    raise typer.Exit(0)



import os
import pathlib
from typing import Optional

from thestage.helpers.error_handler import error_handler
from thestage.services.service_factory import ServiceFactory


def get_current_directory() -> pathlib.Path:
    return pathlib.Path.cwd()


@error_handler()
def validate_config_and_get_service_factory(
        working_directory: Optional[str] = None,
) -> ServiceFactory:
    local_path = get_current_directory() if not working_directory else os.path.abspath(working_directory)

    service_factory = ServiceFactory()
    config_provider = service_factory.get_config_provider()

    config = config_provider.get_config()
    config.runtime.working_directory = str(local_path)

    config_provider.update_config(updated_config=config)

    validation_service = service_factory.get_validation_service()

    validation_service.check_token()
    config_provider.save_config()

    return service_factory



import os
import shutil
from pathlib import Path
from typing import Optional, List

import click
import typer
from rich import print
from tabulate import tabulate

from thestage.color_scheme.color_scheme import ColorScheme
from thestage.config.business.config_provider import ConfigProvider
from thestage.connect.business.remote_server_service import RemoteServerService
from thestage.docker_container.communication.docker_container_api_client import DockerContainerApiClient
from thestage.docker_container.dto.container_response import DockerContainerDto
from thestage.docker_container.dto.enum.container_status import DockerContainerStatus
from thestage.exceptions.git_access_exception import GitAccessException
from thestage.git.communication.git_client import GitLocalClient
from thestage.global_dto.enums.yes_no_response import YesOrNoResponse
from thestage.helpers.error_handler import error_handler
from thestage.i18n.translation import __
from thestage.instance.business.mapper.rented_instance_mapper import RentedInstanceMapper
from thestage.project.business.mapper.project_mapper import ProjectMapper
from thestage.project.communication.project_api_client import ProjectApiClient
from thestage.project.dto.delete_project_response import DeleteProjectResponse
from thestage.project.dto.enum.project_status import ProjectStatus
from thestage.project.dto.modify_project_instance_assignment_response import ModifyProjectInstanceAssignmentResponse
from thestage.project.dto.project_config import ProjectConfig
from thestage.project.dto.project_entity import ProjectEntity
from thestage.project.dto.project_response import ProjectDto
from thestage.services.abstract_service import AbstractService
from thestage.services.clients.thestage_api.core.http_client_exception import HttpClientException
from thestage.services.filesystem_service import FileSystemService
from thestage.task.communication.task_api_client import TaskApiClient
from thestage.task.dto.view_response import TaskViewResponse


class ProjectService(AbstractService):
    __docker_container_api_client: DockerContainerApiClient = None
    __project_api_client: ProjectApiClient = None
    __task_api_client: TaskApiClient = None
    __config_provider: ConfigProvider = None

    def __init__(
            self,
            task_api_client: TaskApiClient,
            project_api_client: ProjectApiClient,
            docker_container_api_client: DockerContainerApiClient,
            config_provider: ConfigProvider,
            remote_server_service: RemoteServerService,
            file_system_service: FileSystemService,
            git_local_client: GitLocalClient,
    ):
        self.__docker_container_api_client = docker_container_api_client
        self.__task_api_client = task_api_client
        self.__project_api_client = project_api_client
        self.__remote_server_service = remote_server_service
        self.__file_system_service = file_system_service
        self.__git_local_client = git_local_client
        self.__config_provider = config_provider


    @error_handler()
    def init_project(
            self,
            project_slug: Optional[str] = None,
            project_public_id: Optional[str] = None,
            remote_url: Optional[str] = None,
    ):
        config = self.__config_provider.get_config()
        project: Optional[ProjectDto] = self.__project_api_client.get_project(
            slug=project_slug,
            public_id=project_public_id,
            default_container_public_id=None
        ).project

        if not project:
            typer.echo('Project not found')
            raise typer.Exit(1)

        if remote_url:
            if os.path.exists(config.runtime.working_directory) and os.listdir(config.runtime.working_directory):
                typer.echo("Working directory must be empty to initialize from remote")
                raise typer.Exit(1)

            self.__git_local_client.clone(
                url=remote_url,
                path=config.runtime.working_directory,
                deploy_key_path=None
            )

            git_dir = os.path.join(config.runtime.working_directory, '.git')
            if os.path.exists(git_dir):
                shutil.rmtree(git_dir)

        is_git_folder = self.__git_local_client.is_present_local_git(
            path=config.runtime.working_directory,
        )
        if is_git_folder:
            has_remote = self.__git_local_client.has_remote(
                path=config.runtime.working_directory,
            )
            if has_remote:
                typer.echo(__('Local repository already has a remote configured; aborting initialization'))
                raise typer.Exit(1)

        if not project.git_repository_url:
            typer.echo(__('Project does not have git repository url'))
            raise typer.Exit(1)

        if project.last_commit_hash or project.last_commit_description:
            continue_with_non_empty_repo: YesOrNoResponse = typer.prompt(
                text=__('Remote repository is probably not empty: latest commit is "{commit_description}" (sha: {commit_hash})\nDo you wish to continue?').format(commit_description=project.last_commit_description, commit_hash=project.last_commit_hash),
                show_choices=True,
                default=YesOrNoResponse.YES.value,
                type=click.Choice([r.value for r in YesOrNoResponse]),
                show_default=True,
            )
            if continue_with_non_empty_repo == YesOrNoResponse.NO:
                typer.echo(__('Project init aborted'))
                raise typer.Exit(0)

        deploy_ssh_key = self.__project_api_client.get_project_deploy_ssh_key(
            public_id=project.public_id,
        )

        deploy_key_path = self.__config_provider.save_project_deploy_ssh_key(
            deploy_ssh_key=deploy_ssh_key,
            project_public_id=project.public_id,
        )

        if is_git_folder:
            has_changes = self.__git_local_client.has_changes_with_untracked(
                path=config.runtime.working_directory,
            )
            if has_changes:
                typer.echo(__('Local repository has uncommitted changes or untracked files. Use an empty directory'))
                raise typer.Exit(1)
        else:
            repo = self.__git_local_client.init_repository(
                path=config.runtime.working_directory,
            )

        is_remote_added = self.__git_local_client.add_remote_to_repo(
            path=config.runtime.working_directory,
            remote_url=project.git_repository_url,
            remote_name=project.git_repository_name,
        )
        if not is_remote_added:
            typer.echo(__('Cannot add remote, something wrong'))
            raise typer.Exit(2)

        self.__git_local_client.configure_local_ssh_command(path=config.runtime.working_directory, deploy_key_path=str(deploy_key_path))

        self.__git_local_client.git_fetch(path=config.runtime.working_directory, deploy_key_path=deploy_key_path)

        self.__git_local_client.init_gitignore(path=config.runtime.working_directory)

        self.__git_local_client.git_add_all(repo_path=config.runtime.working_directory)

        project_config = ProjectConfig()
        project_config.public_id = project.public_id
        project_config.slug = project.slug
        project_config.git_repository_url = project.git_repository_url
        project_config.deploy_key_path = str(deploy_key_path)
        self.__config_provider.save_project_config(project_config=project_config)

        typer.echo(__("Project successfully initialized at %path%", {"path": config.runtime.working_directory}))


    @error_handler()
    def clone_project(
            self,
            project_slug: str,
            project_public_id: str
    ):
        config = self.__config_provider.get_config()
        project: Optional[ProjectDto] = self.__project_api_client.get_project(
            slug=project_slug,
            public_id=project_public_id,
            default_container_public_id=None
        ).project

        if not project:
            typer.echo('Project not found')
            raise typer.Exit(1)

        if not self.__file_system_service.is_folder_empty(folder=config.runtime.working_directory, auto_create=True):
            typer.echo(__("Cannot clone: the folder is not empty"))
            raise typer.Exit(1)

        is_git_folder = self.__git_local_client.is_present_local_git(
            path=config.runtime.working_directory,
        )

        if is_git_folder:
            typer.echo(__('Directory already contains a git repository. Cannot clone here'))
            raise typer.Exit(1)

        if not project.git_repository_url:
            typer.echo(__("Unexpected Project error, missing Repository"))
            raise typer.Exit(1)

        deploy_ssh_key = self.__project_api_client.get_project_deploy_ssh_key(public_id=project.public_id)
        deploy_key_path = self.__config_provider.save_project_deploy_ssh_key(deploy_ssh_key=deploy_ssh_key, project_public_id=project.public_id,)

        try:
            self.__git_local_client.clone(
                url=project.git_repository_url,
                path=config.runtime.working_directory,
                deploy_key_path=deploy_key_path
            )
            self.__git_local_client.init_gitignore(path=config.runtime.working_directory)
        except GitAccessException as ex:
            typer.echo(ex.get_message())
            typer.echo(ex.get_dop_message())
            typer.echo(__(
                "Check you email or open this repo url %git_url% and 'Accept invitation'",
                {
                    'git_url': ex.get_url()
                }
            ))
            raise typer.Exit(1)

        self.__git_local_client.configure_local_ssh_command(path=config.runtime.working_directory, deploy_key_path=str(deploy_key_path))
        project_config = ProjectConfig()
        project_config.public_id = project.public_id
        project_config.slug = project.slug
        project_config.git_repository_url = project.git_repository_url
        project_config.deploy_key_path = str(deploy_key_path)
        self.__config_provider.save_project_config(project_config=project_config)
        typer.echo(__("Project successfully cloned to %path%", {"path": config.runtime.working_directory}))

    @error_handler()
    def checkout_project(
            self,
            task_public_id: Optional[str],
            branch_name: Optional[str],
    ):
        config = self.__config_provider.get_config()
        project_config: ProjectConfig = self.get_fixed_project_config(do_check_default_container=False)
        if not project_config:
            typer.echo(__("Command must be run from an initialized project directory"))
            raise typer.Exit(1)

        target_commit_hash: Optional[str] = None
        if task_public_id:
            task_view_response: Optional[TaskViewResponse] = None
            try:
                task_view_response = self.__task_api_client.get_task(task_public_id=task_public_id)
            except HttpClientException as e:
                if e.get_status_code() == 400:
                    typer.echo(f"Task {task_public_id} was not found")
                    # overriding arguments here
                    branch_name = str(task_public_id)
                    task_public_id = None

            if task_view_response and task_view_response.task:
                target_commit_hash = task_view_response.task.commit_hash
                if not target_commit_hash:
                    typer.echo(f"Task ({task_public_id}) has no commit hash")  # possible legacy problems
                    raise typer.Exit(1)

        is_commit_allowed: bool = True

        if self.__git_local_client.is_head_detached(path=config.runtime.working_directory):
            is_commit_allowed = False
            if self.__git_local_client.is_head_committed_in_headless_state(path=config.runtime.working_directory):
                commit_message = self.__git_local_client.get_current_commit(
                    path=config.runtime.working_directory).message
                print(
                    f"[{ColorScheme.GIT_HEADLESS.value}]Your current commit '{commit_message.strip()}' was likely created in detached head state. Checking out will discard all changes.[/{ColorScheme.GIT_HEADLESS.value}]")
                response: YesOrNoResponse = typer.prompt(
                    text=__('Continue?'),
                    show_choices=True,
                    default=YesOrNoResponse.YES.value,
                    type=click.Choice([r.value for r in YesOrNoResponse]),
                    show_default=True,
                )
                if response == YesOrNoResponse.NO:
                    raise typer.Exit(0)
        else:
            if self.__git_local_client.get_active_branch_name(path=config.runtime.working_directory) == branch_name:
                typer.echo(f"You are already at branch '{branch_name}'")
                raise typer.Exit(0)

        if is_commit_allowed:
            self.__git_local_client.git_add_all(repo_path=config.runtime.working_directory)

            has_changes = self.__git_local_client.has_changes_with_untracked(
                path=config.runtime.working_directory,
            )

            if has_changes:
                active_branch_name = self.__git_local_client.get_active_branch_name(config.runtime.working_directory)
                diff_stat = self.__git_local_client.git_diff_stat(repo_path=config.runtime.working_directory)
                typer.echo(__('Active branch [%branch_name%] has uncommitted changes: %diff_stat_bottomline%', {
                    'diff_stat_bottomline': diff_stat,
                    'branch_name': active_branch_name,
                }))

                response: str = typer.prompt(
                    text=__('Commit changes?'),
                    show_choices=True,
                    default=YesOrNoResponse.YES.value,
                    type=click.Choice([r.value for r in YesOrNoResponse]),
                    show_default=True,
                )
                if response == YesOrNoResponse.NO.value:
                    typer.echo(__('Cannot checkout with uncommitted changes'))
                    raise typer.Exit(0)

                commit_name = typer.prompt(
                    text=__('Provide commit message'),
                    show_choices=False,
                    type=str,
                    show_default=False,
                )

                if commit_name:
                    commit_result = self.__git_local_client.commit_local_changes(
                        path=config.runtime.working_directory,
                        name=commit_name
                    )

                    if commit_result:
                        # in docs not Commit object, on real - str
                        if isinstance(commit_result, str):
                            typer.echo(commit_result)

                    self.__git_local_client.push_changes(
                        path=config.runtime.working_directory,
                        deploy_key_path=project_config.deploy_key_path
                    )
                    typer.echo(__("Pushed changes to remote repository"))
                else:
                    typer.echo(__('Cannot commit with empty commit name'))
                    raise typer.Exit(0)

        if target_commit_hash:
            if self.__git_local_client.get_current_commit(
                    path=config.runtime.working_directory).hexsha != target_commit_hash:
                is_checkout_successful = self.__git_local_client.git_checkout_to_commit(
                    path=config.runtime.working_directory,
                    commit_hash=target_commit_hash
                )

                if is_checkout_successful:
                    print(f"Checked out to commit {target_commit_hash}")
                    print(
                        f"[{ColorScheme.GIT_HEADLESS.value}]HEAD is detached. To be able make changes in repository, checkout to any branch.[/{ColorScheme.GIT_HEADLESS.value}]")
            else:
                typer.echo("HEAD is already at requested commit")
        elif branch_name:
            if self.__git_local_client.is_branch_exists(path=config.runtime.working_directory, branch_name=branch_name):
                self.__git_local_client.git_checkout_to_branch(
                    path=config.runtime.working_directory,
                    branch=branch_name
                )
                typer.echo(f"Checked out to branch '{branch_name}'")
            else:
                typer.echo(f"Branch '{branch_name}' was not found in project repository")
        else:
            main_branch = self.__git_local_client.find_main_branch_name(path=config.runtime.working_directory)
            if main_branch:
                self.__git_local_client.git_checkout_to_branch(
                    path=config.runtime.working_directory,
                    branch=main_branch
                )
                typer.echo(f"Checked out to detected main branch: '{main_branch}'")
            else:
                typer.echo("No main branch found")

    @error_handler()
    def set_default_container(
            self,
            container_public_id: Optional[str],
            container_slug: Optional[str],
    ):
        project_config: ProjectConfig = self.__config_provider.read_project_config()

        if project_config is None:
            typer.echo(f"No project found in working directory")
            raise typer.Exit(1)

        container: Optional[DockerContainerDto] = None
        if container_slug or container_public_id:
            container: DockerContainerDto = self.__docker_container_api_client.get_container(
                container_public_id=container_public_id,
                container_slug=container_slug,
            )
            if container is None:
                typer.echo(f"Could not find container '{container_slug or container_public_id}'")
                raise typer.Exit(1)

            if container.project.public_id != project_config.public_id:
                typer.echo(
                    f"Provided container '{container_slug or container_public_id}' is not related to current project '{project_config.public_id}'")
                raise typer.Exit(1)

            if container.frontend_status.status_key != DockerContainerStatus.RUNNING:
                typer.echo(
                    f"Note: provided container '{container_slug or container_public_id}' is in status '{container.frontend_status.status_translation}'")

        project_config.default_container_public_id = container.public_id if container else None
        project_config.prompt_for_default_container = False
        self.__config_provider.save_project_config(project_config=project_config)
        typer.echo("Default container settings were updated")

    @error_handler()
    def print_project_config(self):
        project_config: ProjectConfig = self.__config_provider.read_project_config()

        if project_config is None:
            typer.echo(f"No project found in working directory")
            raise typer.Exit(1)

        is_deploy_key_exists = project_config.deploy_key_path and self.__file_system_service.check_if_path_exist(
            project_config.deploy_key_path)

        typer.echo(tabulate(
            [
                [
                    "Project ID", project_config.public_id
                ],
                [
                    "Project name", project_config.slug
                ],
                [
                    "Default docker container ID",
                    project_config.default_container_public_id if project_config.default_container_public_id else "<None>"
                ],
                [
                    "Deploy key path", project_config.deploy_key_path if is_deploy_key_exists else "<None>"
                ],
            ],
            showindex=False,
            tablefmt="simple",
        ))

        if is_deploy_key_exists:
            typer.echo("")
            print(f"[{ColorScheme.USEFUL_INFO.value}]Git is configured to use this deploy key automatically.[{ColorScheme.USEFUL_INFO.value}]")
            print(f"[{ColorScheme.USEFUL_INFO.value}]You can run native git commands (e.g. 'git pull', 'git push') directly from project directory.[{ColorScheme.USEFUL_INFO.value}]")

    @error_handler()
    def get_fixed_project_config(self, do_check_default_container: bool) -> Optional[ProjectConfig]:
        project_config: ProjectConfig = self.__config_provider.read_project_config()
        if project_config is None:
            return None

        if project_config.public_id is None or (do_check_default_container and project_config.default_container_public_id):
            project_view_result = self.__project_api_client.get_project(public_id=None, slug=project_config.slug, default_container_public_id=project_config.default_container_public_id)

            if project_config.public_id is None:
                project = project_view_result.project
                project_config.public_id = project.public_id
                self.__config_provider.save_project_config(project_config=project_config)

            if project_config.default_container_public_id and project_view_result.isDefaultContainerDeletedPermanently:
                typer.echo(f"Default container '{project_config.default_container_public_id}' not found, default container setting is cleared")
                project_config.default_container_public_id = None
                project_config.prompt_for_default_container = True
            self.__config_provider.save_project_config(project_config=project_config)

        if not Path(project_config.deploy_key_path).is_file():
            deploy_ssh_key = self.__project_api_client.get_project_deploy_ssh_key(
                public_id=project_config.public_id,
            )

            deploy_key_path = self.__config_provider.save_project_deploy_ssh_key(
                deploy_ssh_key=deploy_ssh_key,
                project_public_id=project_config.public_id,
            )

            project_config.deploy_key_path = deploy_key_path
            self.__config_provider.save_project_config(project_config=project_config)
            typer.echo(f'Recreated missing deploy key for the project')

        return project_config

    @error_handler()
    def pull_project(self):
        config = self.__config_provider.get_config()
        project_config: ProjectConfig = self.get_fixed_project_config(do_check_default_container=False)
        if not project_config:
            typer.echo(__("No project found at the path: %path%. Initialize or clone a project first.",
                          {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        typer.echo("Pulling code from remote repository...")
        self.__git_local_client.git_pull(
            path=config.runtime.working_directory,
            deploy_key_path=project_config.deploy_key_path,
        )

    @error_handler()
    def reset_project(self):
        config = self.__config_provider.get_config()
        project_config: ProjectConfig = self.get_fixed_project_config(do_check_default_container=False)
        if not project_config:
            typer.echo(__("No project found at the path: %path%. Initialize or clone a project first.",
                          {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        typer.echo("Fetching code from remote repository...")
        self.__git_local_client.git_fetch(
            path=config.runtime.working_directory,
            deploy_key_path=project_config.deploy_key_path,
        )
        typer.echo("Resetting local branch...")
        self.__git_local_client.reset_hard(
            path=config.runtime.working_directory,
            deploy_key_path=project_config.deploy_key_path,
            reset_to_origin=True
        )


    @error_handler()
    def modify_project_instance_assignment(
            self,
            project_public_id: Optional[str]= None,
            project_slug: Optional[str]= None,

            rented_instance_public_id_to_assign: Optional[List[str]]= None,
            rented_instance_slug_to_assign: Optional[List[str]]= None,
            rented_instance_public_id_to_delete: Optional[List[str]]= None,
            rented_instance_slug_to_delete: Optional[List[str]]= None,

            self_hosted_instance_public_id_to_assign: Optional[List[str]]= None,
            self_hosted_instance_slug_to_assign: Optional[List[str]]= None,
            self_hosted_instance_public_id_to_delete: Optional[List[str]]= None,
            self_hosted_instance_slug_to_delete: Optional[List[str]]= None,
    ):
        if not project_public_id and not project_slug:
            project_config: ProjectConfig = self.__config_provider.read_project_config()
            if not project_config:
                typer.echo(
                    "Provide the project unique ID or name, or run this command from within an initialized project directory")
                raise typer.Exit(1)
            project_public_id = project_config.public_id

        response: ModifyProjectInstanceAssignmentResponse = self.__project_api_client.modify_project_instance_assignment(
            project_public_id=project_public_id,
            project_slug=project_slug,
            rented_instance_public_id_to_assign=rented_instance_public_id_to_assign or [],
            rented_instance_slug_to_assign=rented_instance_slug_to_assign or [],
            rented_instance_public_id_to_delete=rented_instance_public_id_to_delete or [],
            rented_instance_slug_to_delete=rented_instance_slug_to_delete or [],
            self_hosted_instance_public_id_to_assign=self_hosted_instance_public_id_to_assign or [],
            self_hosted_instance_slug_to_assign=self_hosted_instance_slug_to_assign or [],
            self_hosted_instance_public_id_to_delete=self_hosted_instance_public_id_to_delete or [],
            self_hosted_instance_slug_to_delete=self_hosted_instance_slug_to_delete or [],
        )
        if response:
            if response.isSuccess:
                typer.echo("Project instance assignments were updated")
            else:
                typer.echo(f'Failed to update project instance assignments: {response.message}')
                raise typer.Exit(1)
        else:
            typer.echo("Failed to update project instance assignments")
            raise typer.Exit(1)


    @error_handler()
    def delete_project(self, project_public_id, project_slug):
        if not project_public_id and not project_slug:
            project_config: ProjectConfig = self.__config_provider.read_project_config()
            if not project_config:
                typer.echo(
                    "Provide the project unique ID or name, or run this command from within an initialized project directory")
                raise typer.Exit(1)
            project_public_id = project_config.public_id

        response: DeleteProjectResponse = self.__project_api_client.delete_project(
            project_public_id=project_public_id,
            project_slug=project_slug,
        )
        if response:
            if response.isSuccess:
                typer.echo("Project deleted successfully")
            else:
                typer.echo(f'Failed to delete project: {response.message}')
                raise typer.Exit(1)
        else:
            typer.echo("Failed to delete project")
            raise typer.Exit(1)

    def __get_list(
            self,
            row: int = 5,
            page: int = 1,
            statuses: List[str] = None,
            rented_instance_public_id: Optional[str] = None,
            rented_instance_slug: Optional[str] = None,
            self_hosted_instance_public_id: Optional[str] = None,
            self_hosted_instance_slug: Optional[str] = None,
    ):
        return self.__project_api_client.get_project_list(
            limit=row,
            page=page,
            statuses=statuses,
            rented_instance_public_id=rented_instance_public_id,
            rented_instance_slug=rented_instance_slug,
            self_hosted_instance_public_id=self_hosted_instance_public_id,
            self_hosted_instance_slug=self_hosted_instance_slug,
        )

    @error_handler()
    def print_project_list(
            self,
            row: int,
            page: int,
            statuses: List[str],
            rented_instance_public_id: Optional[str],
            rented_instance_slug: Optional[str],
            self_hosted_instance_public_id: Optional[str],
            self_hosted_instance_slug: Optional[str]
    ):

        status_map = self.__project_api_client.get_project_business_status_map()

        if not statuses:
            default_keys = [ProjectStatus.CREATED]
            statuses = [status_map[k] for k in default_keys if k in status_map]

        if "all" in statuses:
            statuses = list(status_map.values())

        for input_status in statuses:
            if input_status not in status_map.values():
                typer.echo(__("'%invalid_status%' is not one of %valid_statuses%", {
                    'invalid_status': input_status,
                    'valid_statuses': str(list(status_map.values()))
                }))
                raise typer.Exit(1)

        typer.echo(__(
            "Listing projects with statuses: %statuses%. To view all, use --status all",
            placeholders={'statuses': ', '.join(statuses)}
        ))

        backend_statuses = [key for key, value in status_map.items() if value in statuses]
        self.print(
            func_get_data=self.__get_list,
            func_special_params={
                'statuses': backend_statuses,
                'rented_instance_public_id': rented_instance_public_id,
                'rented_instance_slug': rented_instance_slug,
                'self_hosted_instance_public_id': self_hosted_instance_public_id,
                'self_hosted_instance_slug': self_hosted_instance_slug,
            },
            mapper=ProjectMapper(),
            headers=list(map(lambda x: x.alias, ProjectEntity.model_fields.values())),
            row=row,
            page=page,
            max_col_width=[25, 20, 40, 40],
            show_index="never",
        )



from typing import Optional

from pydantic import Field, ConfigDict, BaseModel

from thestage.inference_simulator.dto.inference_simulator import InferenceSimulator


class GetInferenceSimulatorResponse(BaseModel):
    model_config = ConfigDict(use_enum_values=True)
    inferenceSimulator: Optional[InferenceSimulator] = Field(None, alias='inferenceSimulator')





from typing import Optional, List

import typer
from typing_extensions import Annotated

from thestage.cli_command import CliCommand
from thestage.cli_command_helper import get_command_metadata, check_command_permission
from thestage.controllers.utils_controller import validate_config_and_get_service_factory, get_current_directory
from thestage.helpers.logger.app_logger import app_logger
from thestage.i18n.translation import __
from thestage.logging.business.logging_service import LoggingService
from thestage.task.dto.task import Task

app = typer.Typer(no_args_is_help=True, help=__("Manage project tasks"))
runner_app = typer.Typer(help="Project Runner")

@runner_app.command(name='run', no_args_is_help=True, help=__("Run a task within the project. By default, it uses the latest commit from the main branch and streams real-time task logs."), **get_command_metadata(CliCommand.PROJECT_RUN))
def run(
        command: Annotated[List[str], typer.Argument(
            help=__("Command to run (required)"),
        )],
        commit_hash: Optional[str] = typer.Option(
            None,
            '--commit-hash',
            '-hash',
            help=__("Commit hash to use. By default, the current HEAD commit is used."),
            is_eager=False,
        ),
        docker_container_public_id: Optional[str] = typer.Option(
            None,
            '--container-id',
            '-cid',
            help=__("Docker container ID"),
            is_eager=False,
        ),
        docker_container_slug: Optional[str] = typer.Option(
            None,
            '--container-name',
            '-cn',
            help=__("Docker container name"),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            show_default=False,
            is_eager=False,
        ),
        enable_log_stream: Optional[bool] = typer.Option(
            True,
            " /--no-logs",
            " /-nl",
            help=__("Disable real-time log streaming"),
            is_eager=False,
        ),
        task_title: Optional[str] = typer.Option(
            None,
            "--title",
            "-t",
            help=__("Provide a custom task title. Git commit message is used by default."),
            is_eager=False,
        ),
        files_to_add: Optional[str] = typer.Option(
            None,
            "--files-add",
            "-fa",
            help=__("Files to add to the commit. You can add files by their relative path from the working directory with a comma as a separator."),
            is_eager=False,
        ),
        is_skip_auto_commit: Optional[bool] = typer.Option(
            False,
            "--skip-autocommit",
            "-sa",
            help=__("Skip automatic commit of the changes"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_RUN
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [docker_container_public_id, docker_container_slug]) > 1:
        typer.echo("Provide a single identifier for the container - name or ID.")
        raise typer.Exit(1)

    if not command:
        typer.echo(__('Command is required'))
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    task_service = service_factory.get_task_service()

    task: Optional[Task] = task_service.project_run_task(
        run_command=" ".join(command),
        commit_hash=commit_hash,
        docker_container_public_id=docker_container_public_id,
        docker_container_slug=docker_container_slug,
        task_title=task_title,
        files_to_add=files_to_add,
        is_skip_auto_commit=is_skip_auto_commit,
    )

    if enable_log_stream:
        logging_service: LoggingService = service_factory.get_logging_service()
        logging_service.stream_task_logs_with_controls(task_public_id=task.public_id)

    raise typer.Exit(0)


@app.command(name='cancel', no_args_is_help=True, help=__("Cancel a task by ID"), **get_command_metadata(CliCommand.PROJECT_TASK_CANCEL))
def cancel_task(
        task_id: Annotated[str, typer.Argument(
            help=__("Task ID (required)"),
        )],
):
    command_name = CliCommand.PROJECT_TASK_CANCEL
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if not task_id:
        typer.echo('Task ID is required')
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    task_service = service_factory.get_task_service()

    task_service.cancel_task(
        task_public_id=task_id
    )

    raise typer.Exit(0)


@app.command("ls", help=__("List tasks"), **get_command_metadata(CliCommand.PROJECT_TASK_LS))
def list_runs(
        project_public_id: Optional[str] = typer.Option(
            None,
            '--project-id',
            '-pid',
            help=__("Project ID. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        project_slug: Optional[str] = typer.Option(
            None,
            '--project-name',
            '-pn',
            help=__("Project name. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_TASK_LS
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [project_public_id, project_slug]) > 1:
        typer.echo("Provide a single identifier for project - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    task_service = service_factory.get_task_service()

    task_service.print_task_list(project_public_id=project_public_id, project_slug=project_slug, row=row, page=page)

    typer.echo(__("Tasks listing complete"))
    raise typer.Exit(0)


@app.command(name="logs", no_args_is_help=True, help=__("Stream real-time task logs or view last logs for a task"), **get_command_metadata(CliCommand.PROJECT_TASK_LOGS))
def task_logs(
        task_id: Optional[str] = typer.Argument(help=__("Task ID"),),
        logs_number: Optional[int] = typer.Option(
            None,
            '--number',
            '-n',
            help=__("Display a number of latest log entries. No real-time stream if provided."),
            is_eager=False,
        ),
):
    command_name = CliCommand.PROJECT_TASK_LOGS
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if not task_id:
        typer.echo(__('Task ID is required'))
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    logging_service: LoggingService = service_factory.get_logging_service()

    if logs_number is None:
        logging_service.stream_task_logs_with_controls(task_public_id=task_id)
    else:
        logging_service.print_last_task_logs(task_public_id=task_id, logs_number=logs_number)

    app_logger.info(f'Task logs - end')
    raise typer.Exit(0)


import typer

import json
import os
from pathlib import Path
from typing import Optional, Dict, Any

from thestage.config.env_base import THESTAGE_AUTH_TOKEN, THESTAGE_API_URL, THESTAGE_CONFIG_DIR, \
    THESTAGE_CONFIG_FILE
from thestage.cli_command import CliCommand, CliCommandAvailability, ALWAYS_AVAILABLE_COMMANDS
from thestage.services.clients.thestage_api.dtos.validate_token_response import ValidateTokenResponse
from thestage.config.dto.config_entity import ConfigEntity, CliCommandMetadata
from thestage.helpers.ssh_util import parse_private_key
from thestage.config.dto.remote_server_config import RemoteServerConfig
from thestage.services.filesystem_service import FileSystemService
from thestage.project.dto.project_config import ProjectConfig
from thestage.config.business import config_storage


class ConfigProvider:
    _file_system_service = FileSystemService


    def __init__(
            self,
            file_system_service: FileSystemService,
    ):
        self._file_system_service = file_system_service

        global_config_path = self.get_global_config_path()
        if not global_config_path.exists():
            self._file_system_service.create_if_not_exists_dir(global_config_path)

        global_config_file = self.get_global_config_file()
        self._file_system_service.create_if_not_exists_file(global_config_file)


    # must be called right after the app is started
    def build_config(self) -> ConfigEntity:
        config_values = {}

        config_from_env = {}
        config_from_env['main'] = {}
        if THESTAGE_AUTH_TOKEN:
            config_from_env['main']['thestage_auth_token'] = THESTAGE_AUTH_TOKEN
        if THESTAGE_API_URL:
            config_from_env['main']['thestage_api_url'] = THESTAGE_API_URL

        if config_from_env:
            self.__update_config_values_dict(values_to_update=config_values, new_values=config_from_env)

        config_from_file = self._file_system_service.read_config_file(self.get_global_config_file())
        if config_from_file:
            self.__update_config_values_dict(values_to_update=config_values, new_values=config_from_file)

        config = ConfigEntity.model_validate(config_values)
        config.runtime.config_global_path = str(self.get_global_config_path())

        config_storage.APP_CONFIG = config

        return config


    def get_config(self) -> ConfigEntity:
        if not config_storage.APP_CONFIG:
            raise Exception('Config storage was not initialized')
        return config_storage.APP_CONFIG


    def save_project_config(self, project_config: ProjectConfig):
        project_data_dirpath = self.__get_project_config_path(with_file=False)
        self._file_system_service.create_if_not_exists_dir(project_data_dirpath)

        project_data_filepath = self.__get_project_config_path(with_file=True)
        self._file_system_service.create_if_not_exists_file(project_data_filepath)

        self.__save_config_file(data=project_config.model_dump(), file_path=project_data_filepath)


    def save_project_deploy_ssh_key(self, deploy_ssh_key: str, project_public_id: str) -> str:
        deploy_key_dirpath = self.get_global_config_path().joinpath('project_deploy_keys')
        self._file_system_service.create_if_not_exists_dir(deploy_key_dirpath)

        deploy_key_filepath = deploy_key_dirpath.joinpath(f'project_deploy_key__{project_public_id}')
        self._file_system_service.create_if_not_exists_file(deploy_key_filepath)

        text_file = open(deploy_key_filepath, "w")
        text_file.write(deploy_ssh_key)
        text_file.close()
        os.chmod(deploy_key_filepath, 0o600)

        return str(deploy_key_filepath)


    def read_project_config(self) -> Optional[ProjectConfig]:
        project_data_dirpath = self.__get_project_config_path()
        if not project_data_dirpath.exists():
            return None

        project_data_filepath = self.__get_project_config_path(with_file=True)
        if not project_data_filepath.exists():
            return None

        config_data = self._file_system_service.read_config_file(project_data_filepath) if project_data_filepath and project_data_filepath.exists() else {}
        return ProjectConfig.model_validate(config_data)


    def __get_project_config_path(self, with_file: bool = False) -> Path:
        if with_file:
            return Path(self.get_config().runtime.working_directory).joinpath(THESTAGE_CONFIG_DIR).joinpath('project.json')
        else:
            return Path(self.get_config().runtime.working_directory).joinpath(THESTAGE_CONFIG_DIR)


    def __get_remote_server_config_path(self) -> Path:
        return self.get_global_config_path().joinpath('remote_server_config.json')


    def update_remote_server_config_entry(self, ip_address: str, ssh_key_path: Optional[Path]):
        # double checking if key is parseable (fails if key is encrypted)
        # see remote_server_service for similar check
        if ssh_key_path:
            pkey = parse_private_key(str(ssh_key_path))

            if pkey is None:
                typer.echo("Could not identify provided private key (expected RSA, ECDSA, ED25519)")
                raise typer.Exit(1)

        config = self.__read_remote_server_config()
        remote_server_config_filepath = self.__get_remote_server_config_path()
        if config:
            if not config.ip_address_to_ssh_key_map:
                config.ip_address_to_ssh_key_map = {}

            existing_path = config.ip_address_to_ssh_key_map.get(ip_address)
            if ssh_key_path and existing_path != str(ssh_key_path):
                config.ip_address_to_ssh_key_map.update({ip_address: str(ssh_key_path)})
                typer.echo(f'Updated ssh key for {ip_address}: {ssh_key_path}')

            if not ssh_key_path and existing_path:
                typer.echo(f"Private key at path {existing_path} was not found")
                config.ip_address_to_ssh_key_map.pop(ip_address, None)
        else:
            self._file_system_service.create_if_not_exists_file(remote_server_config_filepath)
            config = RemoteServerConfig(ip_address_to_ssh_key_map={ip_address: str(ssh_key_path)})

        self.__save_config_file(data=config.model_dump(), file_path=remote_server_config_filepath)

        return str(remote_server_config_filepath)


    def __read_remote_server_config(self) -> Optional[RemoteServerConfig]:
        config_filepath = self.__get_remote_server_config_path()
        if not config_filepath.is_file():
            return None

        config_data = self._file_system_service.read_config_file(config_filepath) if config_filepath and config_filepath.exists() else {}
        return RemoteServerConfig.model_validate(config_data)


    def get_valid_private_key_path_by_ip_address(self, ip_address: str) -> Optional[str]:
        remote_server_config = self.__read_remote_server_config()
        if remote_server_config and remote_server_config.ip_address_to_ssh_key_map:
            private_key_path = remote_server_config.ip_address_to_ssh_key_map.get(ip_address)
            if private_key_path and Path(private_key_path).is_file():
                return private_key_path
            elif private_key_path:
                self.update_remote_server_config_entry(ip_address, None)
        return None


    @staticmethod
    def __update_config_values_dict(values_to_update: Dict, new_values: Dict):
        if 'main' in new_values:
            if 'main' in values_to_update:
                values_to_update['main'].update(new_values['main'])
            else:
                values_to_update['main'] = new_values['main']

    @staticmethod
    def __save_config_file(data: Dict, file_path: Path):
        with open(file_path, 'w') as configfile:
            json.dump(data, configfile, indent=1)


    def save_config(self):
        config = self.get_config()
        global_config_file = self.get_global_config_file()
        data: Dict[str, Any] = self._file_system_service.read_config_file(global_config_file)
        data.update(config.model_dump(exclude_none=True, by_alias=True, exclude={'runtime', 'RUNTIME', 'daemon', 'DAEMON'}))
        self.__save_config_file(data=data, file_path=global_config_file)


    def clear_config(self, ):
        global_config_path = self.get_global_config_path()
        if global_config_path.exists():
            self._file_system_service.remove_folder(str(global_config_path))

        os.unsetenv('THESTAGE_CONFIG_DIR')
        os.unsetenv('THESTAGE_CONFIG_FILE')
        os.unsetenv('THESTAGE_CLI_ENV')
        os.unsetenv('THESTAGE_API_URL')
        os.unsetenv('THESTAGE_LOG_FILE')
        os.unsetenv('THESTAGE_AUTH_TOKEN')


    def update_config(self, updated_config: ConfigEntity):
        config_storage.APP_CONFIG = updated_config

    def get_global_config_path(self) -> Path:
        home_dir = self._file_system_service.get_home_path()
        return home_dir.joinpath(THESTAGE_CONFIG_DIR)

    def get_global_config_file(self) -> Path:
        return self.get_global_config_path().joinpath(THESTAGE_CONFIG_FILE)


    def update_allowed_commands_and_is_token_valid(self, validate_token_response: Optional[ValidateTokenResponse]):
        config = self.get_config()
        for cli_command_item in CliCommand:
            is_command_available = validate_token_response.allowedCliCommands.get(cli_command_item) if validate_token_response else False
            command_availability = CliCommandAvailability.DEPRECATED

            if cli_command_item in ALWAYS_AVAILABLE_COMMANDS:
                command_availability = CliCommandAvailability.ALLOWED
            else:
                if is_command_available is True:
                    command_availability = CliCommandAvailability.ALLOWED
                if is_command_available is False:
                    command_availability = CliCommandAvailability.RESTRICTED
            config.runtime.allowed_commands.update({cli_command_item: command_availability})
            config.runtime.is_token_valid = validate_token_response.isSuccess if validate_token_response else False
            config.runtime.is_backend_reachable = True if validate_token_response else False
            config.runtime.command_metadata = validate_token_response.cliCommandMetadata if validate_token_response else None

        self.update_config(config)


    def is_cli_command_allowed(self, cli_command: CliCommand) -> bool:
        config = self.get_config()
        return True if config.runtime.allowed_commands.get(cli_command) == CliCommandAvailability.ALLOWED else False



import json
import os
from pathlib import Path
from typing import Optional, List

import click
import typer
from git import Commit
from rich import print

from thestage.color_scheme.color_scheme import ColorScheme
from thestage.config.business.config_provider import ConfigProvider
from thestage.git.communication.git_client import GitLocalClient
from thestage.global_dto.enums.yes_no_response import YesOrNoResponse
from thestage.helpers.error_handler import error_handler
from thestage.i18n.translation import __
from thestage.inference_model.communication.inference_model_api_client import InferenceModelApiClient
from thestage.inference_model.dto.push_inference_simulator_model_response import \
    PushInferenceSimulatorModelResponse
from thestage.inference_simulator.business.mapper.inference_simulator_mapper import InferenceSimulatorMapper
from thestage.inference_simulator.communication.inference_simulator_api_client import InferenceSimulatorApiClient
from thestage.inference_simulator.dto.enum.inference_simulator_status import InferenceSimulatorStatus
from thestage.inference_simulator.dto.get_inference_simulator_response import \
    GetInferenceSimulatorResponse
from thestage.inference_simulator.dto.inference_simulator import InferenceSimulator
from thestage.inference_simulator.dto.inference_simulator_entity import InferenceSimulatorEntity
from thestage.inference_simulator.dto.start_inference_simulator_response import \
    StartInferenceSimulatorResponse
from thestage.project.business.project_service import ProjectService
from thestage.project.dto.project_config import ProjectConfig
from thestage.services.abstract_service import AbstractService
from thestage.services.clients.thestage_api.dtos.paginated_entity_list import PaginatedEntityList


class InferenceSimulatorService(AbstractService):
    def __init__(
            self,
            inference_simulator_api_client: InferenceSimulatorApiClient,
            inference_model_api_client: InferenceModelApiClient,
            config_provider: ConfigProvider,
            git_local_client: GitLocalClient,
            project_service: ProjectService,
    ):
        self.__inference_simulator_api_client = inference_simulator_api_client
        self.__inference_model_api_client = inference_model_api_client
        self.__config_provider = config_provider
        self.__git_local_client = git_local_client
        self.__project_service = project_service

    @error_handler()
    def project_run_inference_simulator(
            self,
            commit_hash: Optional[str] = None,
            rented_instance_public_id: Optional[str] = None,
            rented_instance_slug: Optional[str] = None,
            self_hosted_instance_public_id: Optional[str] = None,
            self_hosted_instance_slug: Optional[str] = None,
            inference_dir: Optional[str] = None,
            is_skip_installation: Optional[bool] = False,
            files_to_add: Optional[str] = None,
            is_skip_auto_commit: Optional[bool] = False,
    ) -> Optional[InferenceSimulator]:
        config = self.__config_provider.get_config()
        project_config: ProjectConfig = self.__project_service.get_fixed_project_config(do_check_default_container=False)

        if not project_config:
            typer.echo(__("No project found at the path: %path%. Initialize or clone a project first. Or provide path to project using --working-directory option.",
                          {"path": config.runtime.working_directory}))
            raise typer.Exit(1)

        instance_args_count = sum(v is not None for v in [rented_instance_public_id, rented_instance_slug, self_hosted_instance_public_id, self_hosted_instance_slug])
        if instance_args_count != 1:
            typer.echo("Provide a single instance (rented or self-hosted) identifier - name or ID.")
            raise typer.Exit(1)

        has_wrong_args = files_to_add and commit_hash or is_skip_auto_commit and commit_hash or files_to_add and is_skip_auto_commit
        if has_wrong_args:
            warning_msg = f"[{ColorScheme.WARNING.value}][WARNING] You can provide only one of the following arguments: --commit-hash, --files-add, --skip-autocommit[{ColorScheme.WARNING.value}]"
            print(warning_msg)
            raise typer.Exit(1)

        if not is_skip_auto_commit and not commit_hash:
            is_git_folder = self.__git_local_client.is_present_local_git(path=config.runtime.working_directory)
            if not is_git_folder:
                typer.echo("Error: Working directory is not a git repository.")
                raise typer.Exit(1)

            is_commit_allowed: bool = True
            has_changes = self.__git_local_client.has_changes_with_untracked(
                path=config.runtime.working_directory,
            )

            if self.__git_local_client.is_head_detached(path=config.runtime.working_directory):
                print(f"[{ColorScheme.GIT_HEADLESS.value}]HEAD is detached[{ColorScheme.GIT_HEADLESS.value}]")

                is_headless_commits_present = self.__git_local_client.is_head_committed_in_headless_state(
                    path=config.runtime.working_directory)
                if is_headless_commits_present:
                    print(
                        f"[{ColorScheme.GIT_HEADLESS.value}]Current commit was made in detached head state. Cannot use it to start the inference simulator. Consider using 'project checkout' command to return to a valid reference.[{ColorScheme.GIT_HEADLESS.value}]")
                    raise typer.Exit(1)

                if has_changes:
                    print(
                        f"[{ColorScheme.GIT_HEADLESS.value}]Local changes detected in detached head state. They will not impact the inference simulator.[{ColorScheme.GIT_HEADLESS.value}]")
                    is_commit_allowed = False
                    response: YesOrNoResponse = typer.prompt(
                        text=__('Continue?'),
                        show_choices=True,
                        default=YesOrNoResponse.YES.value,
                        type=click.Choice([r.value for r in YesOrNoResponse]),
                        show_default=True,
                    )
                    if response == YesOrNoResponse.NO:
                        raise typer.Exit(0)

            if is_commit_allowed:
                if not self.__git_local_client.add_files_with_size_limit_or_warn(config.runtime.working_directory, files_to_add):
                    warning_msg = f"[{ColorScheme.WARNING.value}][WARNING] Inference simulator was not started [{ColorScheme.WARNING.value}]"
                    print(warning_msg)
                    raise typer.Exit(1)

                diff_stat = self.__git_local_client.git_diff_stat(repo_path=config.runtime.working_directory)

                if has_changes and diff_stat:
                    branch_name = self.__git_local_client.get_active_branch_name(config.runtime.working_directory)
                    typer.echo(__('Active branch [%branch_name%] has uncommitted changes: %diff_stat_bottomline%', {
                        'diff_stat_bottomline': diff_stat,
                        'branch_name': branch_name,
                    }))

                    response: str = typer.prompt(
                        text=__('Commit changes?'),
                        show_choices=True,
                        default=YesOrNoResponse.YES.value,
                        type=click.Choice([r.value for r in YesOrNoResponse]),
                        show_default=True,
                    )
                    if response == YesOrNoResponse.NO.value:
                        typer.echo("inference simulator cannot use uncommitted changes - aborting")
                        raise typer.Exit(0)

                    commit_name = typer.prompt(
                        text=__('Provide commit message'),
                        show_choices=False,
                        type=str,
                        show_default=False,
                    )

                    if commit_name:
                        commit_result = self.__git_local_client.commit_local_changes(
                            path=config.runtime.working_directory,
                            name=commit_name
                        )

                        if commit_result:
                            # in docs not Commit object, on real - str
                            if isinstance(commit_result, str):
                                typer.echo(commit_result)

                        self.__git_local_client.push_changes(
                            path=config.runtime.working_directory,
                            deploy_key_path=project_config.deploy_key_path
                        )
                        typer.echo(__("Pushed changes to remote repository"))
                    else:
                        typer.echo(__('Cannot commit with empty commit name, your code will run without last changes.'))
                else:
                    pass
                    # possible to push new empty branch - only that there's a wrong place to do so

        if not commit_hash:
            commit = self.__git_local_client.get_current_commit(path=config.runtime.working_directory)
            if commit and isinstance(commit, Commit):
                commit_hash = commit.hexsha

        start_inference_simulator_response: StartInferenceSimulatorResponse = self.__inference_simulator_api_client.start_project_inference_simulator(
            project_public_id=project_config.public_id,
            commit_hash=commit_hash,
            rented_instance_public_id=rented_instance_public_id,
            rented_instance_slug=rented_instance_slug,
            self_hosted_instance_public_id=self_hosted_instance_public_id,
            self_hosted_instance_slug=self_hosted_instance_slug,
            inference_dir=inference_dir,
            is_skip_installation=is_skip_installation,
        )
        if start_inference_simulator_response:
            if start_inference_simulator_response.message:
                typer.echo(start_inference_simulator_response.message)
            if start_inference_simulator_response.isSuccess and start_inference_simulator_response.inferenceSimulator:
                typer.echo("Inference simulator has been scheduled to run successfully.")
                return start_inference_simulator_response.inferenceSimulator
            else:
                typer.echo(__(
                    'Failed to start inference simulator: %server_massage%',
                    {'server_massage': start_inference_simulator_response.message or ""}
                ))
                raise typer.Exit(1)
        else:
            typer.echo(__("Failed to start inference simulator"))
            raise typer.Exit(1)


    @error_handler()
    def project_push_inference_simulator(
            self,
            public_id: Optional[str] = None,
            slug: Optional[str] = None,
    ):

        push_inference_simulator_model_response: PushInferenceSimulatorModelResponse = self.__inference_model_api_client.push_project_inference_simulator_model(
            public_id=public_id,
            slug=slug,
        )
        if push_inference_simulator_model_response:
            if push_inference_simulator_model_response.message:
                typer.echo(push_inference_simulator_model_response.message)
            if push_inference_simulator_model_response.isSuccess:
                typer.echo("Inference simulator has been successfully scheduled to be pushed to S3 and ECR.")
            else:
                typer.echo(__(
                    'Failed to push inference simulator: %server_massage%',
                    {'server_massage': push_inference_simulator_model_response.message or ""}
                ))
                raise typer.Exit(1)
        else:
            typer.echo(__("Failed to push inference simulator"))
            raise typer.Exit(1)

    @error_handler()
    def project_get_and_save_inference_simulator_metadata(
            self,
            inference_simulator_public_id: Optional[str] = None,
            inference_simulator_slug: Optional[str] = None,
            file_path: Optional[str] = None,
    ):
        get_inference_metadata_response: GetInferenceSimulatorResponse = self.__inference_simulator_api_client.get_inference_simulator(
            public_id=inference_simulator_public_id,
            slug=inference_simulator_slug,
        )

        metadata = get_inference_metadata_response.inferenceSimulator.qlip_serve_metadata

        if metadata:
            typer.echo("qlip_serve_metadata:")
            typer.echo(json.dumps(metadata, indent=4))

            if not file_path:
                file_path = Path(os.getcwd()) / "metadata.json"
                typer.echo(__("No file path provided. Saving metadata to %file_path%", {"file_path": str(file_path)}))

            try:
                parsed_metadata = metadata

                output_file = Path(file_path)
                output_file.parent.mkdir(parents=True, exist_ok=True)
                with output_file.open("w", encoding="utf-8") as file:
                    json.dump(parsed_metadata, file, indent=4)
                typer.echo(__("Metadata successfully saved to %file_path%", {"file_path": str(file_path)}))
            except Exception as e:
                typer.echo(__("Failed to save metadata to %file_path%. Error: %error%",
                              {"file_path": file_path, "error": str(e)}))
                raise typer.Exit(1)
        else:
            typer.echo(__("No qlip_serve_metadata found"))
            raise typer.Exit(1)


    @error_handler()
    def get_project_inference_simulator_list(
            self,
            project_public_id: Optional[str],
            project_slug: Optional[str],
            statuses: List[str],
            row: int = 5,
            page: int = 1,
    ) -> PaginatedEntityList[InferenceSimulator]:
        data: Optional[PaginatedEntityList[InferenceSimulator]] = self.__inference_simulator_api_client.get_inference_simulator_list(
            statuses=statuses,
            project_public_id=project_public_id,
            project_slug=project_slug,
            page=page,
            limit=row,
        )

        return data

    @error_handler()
    def print_inference_simulator_list(self, project_public_id, project_slug, statuses, row, page):
        if not project_public_id and not project_slug:
            project_config: ProjectConfig = self.__config_provider.read_project_config()
            if not project_config:
                typer.echo(__("Provide the project unique ID or run this command from within an initialized project directory"))
                raise typer.Exit(1)
            project_public_id = project_config.public_id

        inference_simulator_status_map = self.__inference_simulator_api_client.get_inference_simulator_business_status_map()

        if not statuses:
            statuses = ({key: inference_simulator_status_map[key] for key in [
                InferenceSimulatorStatus.SCHEDULED,
                InferenceSimulatorStatus.CREATING,
                InferenceSimulatorStatus.RUNNING,
            ]}).values()

        if "all" in statuses:
            statuses = inference_simulator_status_map.values()

        for input_status_item in statuses:
            if input_status_item not in inference_simulator_status_map.values():
                typer.echo(__("'%invalid_status%' is not one of %valid_statuses%", {
                    'invalid_status': input_status_item,
                    'valid_statuses': str(list(inference_simulator_status_map.values()))
                }))
                raise typer.Exit(1)

        typer.echo(__(
            "Listing inference simulators with the following statuses: %statuses%, to view all inference simulators, use --status all",
            placeholders={
                'statuses': ', '.join([status_item for status_item in statuses])
            }))

        backend_statuses: List[str] = [key for key, value in inference_simulator_status_map.items() if value in statuses]

        self.print(
            func_get_data=self.get_project_inference_simulator_list,
            func_special_params={
                'project_public_id': project_public_id,
                'project_slug': project_slug,
                'statuses': backend_statuses,
            },
            mapper=InferenceSimulatorMapper(),
            headers=list(map(lambda x: x.alias, InferenceSimulatorEntity.model_fields.values())),
            row=row,
            page=page,
            max_col_width=[100, 100, 100, 100, 100, 100, 100, 100],
            show_index="never",
        )


from pathlib import Path
from typing import Optional, List

import typer

from thestage.cli_command import CliCommand
from thestage.cli_command_helper import get_command_metadata, check_command_permission
from thestage.controllers.utils_controller import validate_config_and_get_service_factory, get_current_directory
from thestage.docker_container.business.container_service import ContainerService
from thestage.docker_container.dto.container_response import DockerContainerDto
from thestage.docker_container.dto.enum.container_pending_action import DockerContainerAction
from thestage.helpers.logger.app_logger import app_logger
from thestage.i18n.translation import __
from thestage.logging.business.logging_service import LoggingService

app = typer.Typer(no_args_is_help=True, help=__("Manage containers"))

@app.command(name='ls', help=__("List containers"), **get_command_metadata(CliCommand.CONTAINER_LS))
def list_containers(
        row: int = typer.Option(
            5,
            '--row',
            '-r',
            help=__("Set number of rows displayed per page"),
            is_eager=False,
        ),
        page: int = typer.Option(
            1,
            '--page',
            '-p',
            help=__("Set starting page for displaying output"),
            is_eager=False,
        ),
        project_public_id: str = typer.Option(
            None,
            '--project-id',
            '-pid',
            help=__("Filter containers by project, using the project's ID"),
            is_eager=False,
        ),
        project_slug: str = typer.Option(
            None,
            '--project-name',
            '-pn',
            help=__("Filter containers by project, using the project's name"),
            is_eager=False,
        ),
        statuses: List[str] = typer.Option(
            None,
            '--status',
            '-s',
            help=__("Filter by status, use --status all to list all containers"),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONTAINER_LS
    app_logger.info(f'Running {command_name}')
    check_command_permission(command_name)

    if sum(v is not None for v in [project_public_id, project_slug]) > 1:
        typer.echo("Provide a single identifier for project - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    container_service: ContainerService = service_factory.get_container_service()
    container_service.print_container_list(
        row=row,
        page=page,
        project_public_id=project_public_id,
        project_slug=project_slug,
        statuses=statuses,
    )

    typer.echo(__("Containers listing complete"))
    raise typer.Exit(0)


@app.command(name="info", no_args_is_help=True, help=__("Get container info"), **get_command_metadata(CliCommand.CONTAINER_INFO))
def container_info(
        container_public_id: str = typer.Option(
            None,
            '--container-id',
            '-cid',
            help=__("Container ID"),
            is_eager=False,
        ),
        container_slug: str = typer.Option(
            None,
            '--container-name',
            '-cn',
            help=__("Container name"),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONTAINER_INFO
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [container_public_id, container_slug]) != 1:
        typer.echo("Provide a single identifier for container - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    container_service: ContainerService = service_factory.get_container_service()

    container: Optional[DockerContainerDto] = container_service.get_container(
        container_public_id=container_public_id,
        container_slug=container_slug,
    )

    if not container:
        typer.echo("Container not found")
        raise typer.Exit(1)

    typer.echo(__("STATUS: %status%", {'status': str(container.frontend_status.status_translation if container and container.frontend_status else 'UNKNOWN')}))
    typer.echo(__("ID: %slug%", {'slug': str(container.public_id)}))
    typer.echo(__("NAME: %title%", {'title': str(container.slug)}))
    typer.echo(__("IMAGE: %image%", {'image': str(container.docker_image)}))

    typer.echo(f"CONTAINER INSTANCE:")

    if container.instance_rented:
        typer.echo(f"    TYPE: RENTED")
        typer.echo(f"    ID: {container.instance_rented.public_id}")
        typer.echo(f"    NAME: {container.instance_rented.slug}")
        typer.echo(
            __("RENTED INSTANCE STATUS: %instance_status%",
               {'instance_status': str(container.instance_rented.frontend_status.status_translation if container.instance_rented.frontend_status else 'UNKNOWN')})
        )

    if container.selfhosted_instance:
        typer.echo(f"    TYPE: SELF-HOSTED")
        typer.echo(f"    ID: {container.selfhosted_instance.public_id}")
        typer.echo(f"    NAME: {container.selfhosted_instance.slug}")
        typer.echo(
            __("SELF-HOSTED INSTANCE STATUS: %instance_status%",
               {'instance_status': str(container.selfhosted_instance.frontend_status.status_translation if container.selfhosted_instance.frontend_status else 'UNKNOWN')})
        )

    if container.mappings and (container.mappings.port_mappings or container.mappings.directory_mappings):
        if container.mappings.port_mappings:
            typer.echo(__("CONTAINER PORT MAPPING:"))
            for src, dest in container.mappings.port_mappings.items():
                typer.echo(f"    {src} : {dest}")

        if container.mappings.directory_mappings:
            typer.echo(__("CONTAINER DIRECTORY MAPPING:"))
            for src, dest in container.mappings.directory_mappings.items():
                typer.echo(f"    {src} : {dest}")

    raise typer.Exit(0)


@app.command(name="connect", no_args_is_help=True, help=__("Connect to container"), **get_command_metadata(CliCommand.CONTAINER_CONNECT))
def container_connect(
        container_public_id: str = typer.Option(
            None,
            '--container-id',
            '-cid',
            help=__("Container ID"),
            is_eager=False,
        ),
        container_slug: str = typer.Option(
            None,
            '--container-name',
            '-cn',
            help=__("Container name"),
            is_eager=False,
        ),
        username: Optional[str] = typer.Option(
            None,
            '--username',
            '-u',
            help=__("Username for the server instance (required when connecting to self-hosted instance)"),
            is_eager=False,
        ),
        private_ssh_key_path: str = typer.Option(
            None,
            "--private-key-path",
            "-pk",
            help=__("Path to private key that will be accepted by remote server (optional)"),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONTAINER_CONNECT
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [container_public_id, container_slug]) != 1:
        typer.echo("Provide a single identifier for container - ID or name.")
        raise typer.Exit(1)

    if private_ssh_key_path and not Path(private_ssh_key_path).is_file():
        typer.echo(f'No file found at provided path {private_ssh_key_path}')
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    container_service: ContainerService = service_factory.get_container_service()

    container_service.connect_to_container(
        container_public_id=container_public_id,
        container_slug=container_slug,
        username=username,
        input_ssh_key_path=private_ssh_key_path,
    )

    app_logger.info(f'Stop connect to container')
    raise typer.Exit(0)


@app.command(name="upload", no_args_is_help=True, help=__("Upload file to container"), **get_command_metadata(CliCommand.CONTAINER_UPLOAD))
def upload_file(
        source_path: str = typer.Argument(help=__("Source file path"),),
        destination: Optional[str] = typer.Argument(help=__("Destination directory path in container. Format: container_id_or_name:/path/to/file"),),
        username: Optional[str] = typer.Option(
            None,
            '--username',
            '-u',
            help=__("Username for the server instance (required when connecting to self-hosted instance)"),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONTAINER_UPLOAD
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    service_factory = validate_config_and_get_service_factory()
    container_service: ContainerService = service_factory.get_container_service()

    container_service.put_file_to_container(
        source_path=source_path,
        destination=destination,
        username_param=username,
    )

    app_logger.info(f'File upload completed')
    raise typer.Exit(0)


@app.command(name="download", no_args_is_help=True, help=__("Download file from container"), **get_command_metadata(CliCommand.CONTAINER_DOWNLOAD))
def download_file(
        source: str = typer.Argument(help=__("Source file path in container. Format: container_name:/path/to/file"),),
        destination_path: str = typer.Argument(help=__("Destination directory path on local machine"),),
        username: Optional[str] = typer.Option(
            None,
            '--username',
            '-u',
            help=__("Username for the server instance (required when connecting to self-hosted instance)"),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONTAINER_DOWNLOAD
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    service_factory = validate_config_and_get_service_factory()
    container_service: ContainerService = service_factory.get_container_service()

    container_service.get_file_from_container(
        source=source,
        destination_path=destination_path.rstrip("/"),
        username_param=username,
    )

    app_logger.info(f'File download completed')
    raise typer.Exit(0)


@app.command(name="start", no_args_is_help=True, help=__("Start container"), **get_command_metadata(CliCommand.CONTAINER_START))
def start_container(
        container_public_id: str = typer.Option(
            None,
            '--container-id',
            '-cid',
            help=__("Container ID"),
            is_eager=False,
        ),
        container_slug: str = typer.Option(
            None,
            '--container-name',
            '-cn',
            help=__("Container name"),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONTAINER_START
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [container_public_id, container_slug]) != 1:
        typer.echo("Provide a single identifier for container - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    container_service: ContainerService = service_factory.get_container_service()

    container_service.request_docker_container_action(
        container_public_id=container_public_id,
        container_slug=container_slug,
        action=DockerContainerAction.START
    )

    app_logger.info(f'Container is starting')
    raise typer.Exit(0)


@app.command(name="create", no_args_is_help=True, help=__("Create container"), **get_command_metadata(CliCommand.CONTAINER_CREATE))
def create_container(
        project_public_id: Optional[str] = typer.Option(
            None,
            '--project-id',
            '-pid',
            help=__("Project ID. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        project_slug: Optional[str] = typer.Option(
            None,
            '--project-name',
            '-pn',
            help=__("Project name. By default, project info is taken from the current directory"),
            is_eager=False,
        ),
        working_directory: Optional[str] = typer.Option(
            None,
            "--working-directory",
            "-wd",
            help=__("Full path to working directory"),
            show_default=False,
            is_eager=False,
        ),
        container_slug: Optional[str] = typer.Option(
            None,
            '--container-name',
            '-cn',
            help=__("Container name"),
            is_eager=False,
        ),
        rented_instance_public_id: Optional[str] = typer.Option(
            None,
            '--rented-instance-id',
            '-rid',
            help=__("Rented instance ID on which container will run"),
            is_eager=False,
        ),
        rented_instance_slug: Optional[str] = typer.Option(
            None,
            '--rented-instance-name',
            '-rn',
            help=__("Rented instance name on which container will run"),
            is_eager=False,
        ),
        self_hosted_instance_public_id: Optional[str] = typer.Option(
            None,
            '--self-hosted-instance-id',
            '-sid',
            help=__("Self-hosted instance ID on which container will run"),
            is_eager=False,
        ),
        self_hosted_instance_slug: Optional[str] = typer.Option(
            None,
            '--self-hosted-instance-name',
            '-sn',
            help=__("Self-hosted instance name on which container will run"),
            is_eager=False,
        ),
        image_name: Optional[str] = typer.Option(
            None,
            '--image',
            '-i',
            help=__("Docker image name. If not provided, you will be prompted to select from available images"),
            is_eager=False,
        ),
        gpu_number: Optional[int] = typer.Option(
            None,
            '--gpu-number',
            '-gpun',
            help=__("Number of assigned GPUs for container. No GPU limit by default"),
            is_eager=False,
        ),
        cpu_number: Optional[int] = typer.Option(
            None,
            '--cpu-number',
            '-cpun',
            help=__("Number of assigned CPUs for container. No CPU limit by default"),
            is_eager=False,
        ),
        volume_mounts: Optional[List[str]] = typer.Option(
            None,
            '--volume',
            '-v',
            help=__("Volume mounts. Can be used multiple times. Format: /host_dir:/container_dir"),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONTAINER_CREATE
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [project_public_id, project_slug]) > 1:
        typer.echo("Provide a single identifier for project - ID or name.")
        raise typer.Exit(1)

    if sum(v is not None for v in [rented_instance_public_id, rented_instance_slug, self_hosted_instance_public_id, self_hosted_instance_slug]) != 1:
        typer.echo("Provide exactly one identifier (ID or Name) for either a rented or self-hosted instance.")
        raise typer.Exit(1)

    if not container_slug:
        container_slug = typer.prompt("Enter container name")

    volume_mappings_dict = {}
    if volume_mounts:
        for mount in volume_mounts:
            if ":" not in mount:
                typer.echo(f"Invalid volume format: {mount}. Expected host:container")
                raise typer.Exit(1)
            host_path, container_path = mount.split(":", 1)
            volume_mappings_dict[host_path] = container_path

    service_factory = validate_config_and_get_service_factory(working_directory=working_directory)
    container_service: ContainerService = service_factory.get_container_service()

    container_service.create_docker_container(
        project_public_id=project_public_id,
        project_slug=project_slug,
        container_slug=container_slug,
        rented_instance_public_id=rented_instance_public_id,
        rented_instance_slug=rented_instance_slug,
        self_hosted_instance_public_id=self_hosted_instance_public_id,
        self_hosted_instance_slug=self_hosted_instance_slug,
        image_name=image_name,
        cpu_number=cpu_number,
        gpu_number=gpu_number,
        volume_mappings=volume_mappings_dict
    )

    app_logger.info(f'Container is creating')
    raise typer.Exit(0)


@app.command(name="stop", no_args_is_help=True, help=__("Stop container"), **get_command_metadata(CliCommand.CONTAINER_STOP))
def stop_container(
        container_public_id: str = typer.Option(
            None,
            '--container-id',
            '-cid',
            help=__("Container ID"),
            is_eager=False,
        ),
        container_slug: str = typer.Option(
            None,
            '--container-name',
            '-cn',
            help=__("Container name"),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONTAINER_STOP
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [container_public_id, container_slug]) != 1:
        typer.echo("Provide a single identifier for container - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    container_service: ContainerService = service_factory.get_container_service()

    container_service.request_docker_container_action(
        container_public_id=container_public_id,
        container_slug=container_slug,
        action=DockerContainerAction.STOP
    )

    app_logger.info(f'Container is stopping')
    raise typer.Exit(0)


@app.command(name="delete", no_args_is_help=True, help=__("Delete container"), **get_command_metadata(CliCommand.CONTAINER_DELETE))
def delete_container(
        container_public_id: str = typer.Option(
            None,
            '--container-id',
            '-cid',
            help=__("Container ID"),
            is_eager=False,
        ),
        container_slug: str = typer.Option(
            None,
            '--container-name',
            '-cn',
            help=__("Container name"),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONTAINER_DELETE
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [container_public_id, container_slug]) != 1:
        typer.echo("Provide a single identifier for container - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    container_service: ContainerService = service_factory.get_container_service()

    container_service.request_docker_container_action(
        container_public_id=container_public_id,
        container_slug=container_slug,
        action=DockerContainerAction.DELETE
    )

    app_logger.info(f'Container is deleting')
    raise typer.Exit(0)

@app.command(name="restart", no_args_is_help=True, help=__("Restart container"), **get_command_metadata(CliCommand.CONTAINER_RESTART))
def restart_container(
        container_public_id: str = typer.Option(
            None,
            '--container-id',
            '-cid',
            help=__("Container ID"),
            is_eager=False,
        ),
        container_slug: str = typer.Option(
            None,
            '--container-name',
            '-cn',
            help=__("Container name"),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONTAINER_RESTART
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [container_public_id, container_slug]) != 1:
        typer.echo("Provide a single identifier for container - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    container_service: ContainerService = service_factory.get_container_service()

    container_service.request_docker_container_action(
        container_public_id=container_public_id,
        container_slug=container_slug,
        action=DockerContainerAction.RESTART
    )

    app_logger.info(f'Container is restarting')
    raise typer.Exit(0)


@app.command(name="logs", no_args_is_help=True, help=__("Stream real-time Docker container logs or view last logs for a container"), **get_command_metadata(CliCommand.CONTAINER_LOGS))
def container_logs(
        container_public_id: str = typer.Option(
            None,
            '--container-id',
            '-cid',
            help=__("Container ID"),
            is_eager=False,
        ),
        container_slug: str = typer.Option(
            None,
            '--container-name',
            '-cn',
            help=__("Container name"),
            is_eager=False,
        ),
        logs_number: Optional[int] = typer.Option(
            None,
            '--number',
            '-n',
            help=__("Display a number of latest log entries. No real-time stream if provided."),
            is_eager=False,
        ),
):
    command_name = CliCommand.CONTAINER_LOGS
    app_logger.info(f'Running {command_name} from {get_current_directory()}')
    check_command_permission(command_name)

    if sum(v is not None for v in [container_public_id, container_slug]) != 1:
        typer.echo("Provide a single identifier for container - ID or name.")
        raise typer.Exit(1)

    service_factory = validate_config_and_get_service_factory()
    logging_service: LoggingService = service_factory.get_logging_service()

    if logs_number is None:
        logging_service.stream_container_logs_with_controls(
            container_public_id=container_public_id,
            container_slug=container_slug,
        )
    else:
        logging_service.print_last_container_logs(container_public_id=container_public_id, container_slug=container_slug, logs_number=logs_number)

    app_logger.info(f'Container log streaming completed')
    raise typer.Exit(0)


from enum import Enum

# https://rich.readthedocs.io/en/stable/appendix/colors.html
class ColorScheme(str, Enum):
    GIT_HEADLESS = "orange_red1"
    WARNING = "orange_red1"
    USEFUL_INFO = "deep_sky_blue1"
    SUCCESS = "green"
